# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from anaconda_project.internal.cli.variable_commands import main_add, main_remove
from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents_completing_project_file
from anaconda_project.internal.simple_status import SimpleStatus
from anaconda_project.internal import conda_api
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME

PLATFORM_ENV_VAR = conda_api.conda_prefix_variable()


class Args(object):
    def __init__(self, vars_to_add=None, vars_to_remove=None, directory='.', default=None, env_spec=None):
        self.directory = directory
        self.vars_to_add = vars_to_add
        self.vars_to_remove = vars_to_remove
        self.default = None
        self.env_spec = None


def _monkeypatch_add_variables(monkeypatch):
    params = []

    def mock_add_variables(project, env_spec_name, _vars, defaults):
        params.append(env_spec_name)
        params.append(_vars)
        params.append(defaults)
        return True

    monkeypatch.setattr('anaconda_project.project_ops.add_variables', mock_add_variables)

    return params


def test_add_variable_command(monkeypatch):

    params = _monkeypatch_add_variables(monkeypatch)

    args = Args(vars_to_add=['foo', 'baz'])
    res = main_add(args)
    assert res == 0
    assert params[0] is None
    assert ['foo', 'baz'] == params[1]


def test_add_variable_with_default(monkeypatch):

    params = _monkeypatch_add_variables(monkeypatch)

    res = _parse_args_and_run_subcommand(['anaconda-project', 'add-variable', '--default', 'bar', 'foo'])
    assert res == 0
    assert [None, ['foo'], dict(foo='bar')] == params


def test_add_variable_with_env_spec(monkeypatch):

    params = _monkeypatch_add_variables(monkeypatch)

    res = _parse_args_and_run_subcommand(['anaconda-project', 'add-variable', '--env-spec', 'bar', 'foo'])
    assert res == 0
    assert ['bar', ['foo'], dict(foo=None)] == params


def test_add_two_variables_with_default(monkeypatch, capsys):

    params = _monkeypatch_add_variables(monkeypatch)

    res = _parse_args_and_run_subcommand(['anaconda-project', 'add-variable', '--default', 'bar', 'foo', 'hello'])
    assert res == 1
    assert [] == params

    out, err = capsys.readouterr()
    assert out == ''
    expected_err = ("It isn't clear which variable your --default option goes with; " +
                    "add one variable at a time if using --default.\n")
    assert err == expected_err


def test_add_variable_project_problem(capsys):
    def check_problem(dirname):
        args = Args(vars_to_add=['foo', 'baz'], directory=dirname)
        res = main_add(args)
        assert res == 1

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ("variables:\n" "  42")}, check_problem)

    out, err = capsys.readouterr()
    assert out == ''
    expected_err = ('variables section contains wrong value type 42, should be dict or list of requirements\n'
                    'Unable to load the project.\n')
    assert expected_err in err


def _monkeypatch_remove_variables(monkeypatch):
    params = []

    def mock_remove_variables(project, env_spec_name, _vars):
        params.append(env_spec_name)
        params.append(_vars)
        return True

    monkeypatch.setattr('anaconda_project.project_ops.remove_variables', mock_remove_variables)

    return params


def test_remove_variable_command(monkeypatch):

    params = _monkeypatch_remove_variables(monkeypatch)

    def check_remove_variable(dirname):
        args = Args(vars_to_remove=['foo', 'baz'], directory=dirname)
        res = main_remove(args)
        assert res == 0
        assert len(params) == 2

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: ("variables:\n"
                                    "  foo: {default: test}\n"
                                    "  baz: {default: bar}")}, check_remove_variable)


def test_remove_variable_project_problem(monkeypatch):
    def check_problem_remove(dirname):
        args = Args(vars_to_remove=['foo', 'baz'], directory=dirname)
        res = main_remove(args)
        assert res == 1

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: ("variables:\n"
                                    "  foo: true")}, check_problem_remove)


def test_list_variables(capsys):
    def check_list_not_empty(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-variables', '--directory', dirname])

        assert code == 0
        out, err = capsys.readouterr()
        expected_out = """
Variables for project: {dirname}

Name{space}Description
===={space}===========
{varname}  The project needs a Conda environment containing all required packages.
tes2{space}A downloaded file which is referenced by tes2.
test{space}A downloaded file which is referenced by test.
""".format(dirname=dirname,
           varname=PLATFORM_ENV_VAR,
           space="".ljust(len(PLATFORM_ENV_VAR) - 2)).strip() + "\n"
        assert out == expected_out

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: ('downloads:\n'
                                    '  test: http://localhost:8000/test.tgz\n'
                                    '  tes2: http://localhost:8000/train.tgz\n')}, check_list_not_empty)


def test_list_variables_with_no_variables(capsys):
    def check_list_empty(dirname):
        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-variables', '--directory', dirname])

        assert code == 0
        out, err = capsys.readouterr()
        expected_out = """
Variables for project: {dirname}

Name{space}Description
===={space}===========
{varname}  The project needs a Conda environment containing all required packages.
""".format(dirname=dirname,
           varname=PLATFORM_ENV_VAR,
           space="".ljust(len(PLATFORM_ENV_VAR) - 2)).strip() + "\n"
        assert out == expected_out

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check_list_empty)


def test_list_variables_with_project_file_problems(capsys):
    def check(dirname):

        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-variables', '--directory', dirname])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert ('variables section contains wrong value type 42,' + ' should be dict or list of requirements\n' +
                'Unable to load the project.\n') in err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, check)


def test_set_variables_with_project_file_problems(capsys):
    def check(dirname):

        code = _parse_args_and_run_subcommand(['anaconda-project', 'set-variable', '--directory', dirname, 'FOO=bar'])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert ('variables section contains wrong value type 42,' + ' should be dict or list of requirements\n' +
                'Unable to load the project.\n') in err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, check)


def test_unset_variables_with_project_file_problems(capsys):
    def check(dirname):

        code = _parse_args_and_run_subcommand(['anaconda-project', 'unset-variable', '--directory', dirname, 'FOO'])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert ('variables section contains wrong value type 42,' + ' should be dict or list of requirements\n' +
                'Unable to load the project.\n') in err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, check)


def _monkeypatch_set_variables(monkeypatch):
    params = []

    def mock_set_variables(project, env_spec_name, _vars):
        params.append(env_spec_name)
        params.append(_vars)
        return SimpleStatus(success=True, description="BOO")

    monkeypatch.setattr('anaconda_project.project_ops.set_variables', mock_set_variables)

    return params


def test_set_variable_command(monkeypatch):

    params = _monkeypatch_set_variables(monkeypatch)

    def check(dirname):
        res = _parse_args_and_run_subcommand(['anaconda-project', 'set-variable', '--directory', dirname, 'foo=bar',
                                              'baz=qux', 'has_two_equals=foo=bar'])
        assert res == 0

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  - foo
  - baz
  - has_two_equals
    """}, check)

    assert [('foo', 'bar'), ('baz', 'qux'), ('has_two_equals', 'foo=bar')] == params[1]


def test_set_variable_command_with_env_spec(monkeypatch):

    params = _monkeypatch_set_variables(monkeypatch)

    def check(dirname):
        res = _parse_args_and_run_subcommand(['anaconda-project', 'set-variable', '--env-spec', 'somespec',
                                              '--directory', dirname, 'foo=bar', 'baz=qux', 'has_two_equals=foo=bar'])
        assert res == 0

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  - foo
  - baz
  - has_two_equals
    """}, check)

    assert 'somespec' == params[0]
    assert [('foo', 'bar'), ('baz', 'qux'), ('has_two_equals', 'foo=bar')] == params[1]


def test_set_variable_command_bad_arg(monkeypatch, capsys):

    params = _monkeypatch_set_variables(monkeypatch)

    res = _parse_args_and_run_subcommand(['anaconda-project', 'set-variable', 'foo=bar', 'baz'])
    assert res == 1
    out, err = capsys.readouterr()
    assert "Error: argument '{}' should be in NAME=value format".format('baz') in out

    assert len(params) == 0


def _monkeypatch_unset_variables(monkeypatch):
    params = []

    def mock_unset_variables(project, env_spec_name, _vars):
        params.append(env_spec_name)
        params.append(_vars)
        return SimpleStatus(success=True, description="BOO")

    monkeypatch.setattr('anaconda_project.project_ops.unset_variables', mock_unset_variables)

    return params


def test_unset_variable_command(monkeypatch):

    params = _monkeypatch_unset_variables(monkeypatch)

    def check(dirname):
        res = _parse_args_and_run_subcommand(['anaconda-project', 'unset-variable', '--directory', dirname, 'foo', 'baz'
                                              ])
        assert res == 0

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
variables:
  - foo
  - baz
    """}, check)

    assert params[0] is None
    assert ['foo', 'baz'] == params[1]



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Talking to the Anaconda server."""
from __future__ import absolute_import, print_function

import logging
import os
import tarfile
import zipfile

import requests
import binstar_client.utils as binstar_utils
import binstar_client.requests_ext as binstar_requests_ext
from binstar_client.errors import BinstarError, Unauthorized

from anaconda_project.internal.simple_status import SimpleStatus


class _Client(object):
    def __init__(self, site=None, username=None, token=None, log_level=None):
        assert hasattr(binstar_utils, 'get_server_api'), "Please upgrade anaconda-client"
        if log_level is None:
            log_level = logging.INFO
        self._api = binstar_utils.get_server_api(site=site, token=token, log_level=log_level)
        self._user_info = None
        self._force_username = username

    def _username(self):
        """Get username to upload to; raise Unauthorized if we aren't logged in."""
        if self._user_info is None:
            self._user_info = self._api.user(login=self._force_username)
        assert self._user_info is not None
        username = self._user_info.get('login', None)
        if username is None:
            raise Unauthorized()
        else:
            return username

    def _check_response(self, res, allowed=(200, 201)):
        # using a little private API here.
        self._api._check_response(res, allowed=list(allowed))

    # HACK ALERT: using this is a workaround for
    # https://github.com/Anaconda-Platform/anaconda-server/issues/2229
    def _exists(self, project_name):
        url = "{}/apps/{}/projects/{}".format(self._api.domain, self._username(), project_name)
        res = self._api.session.get(url)
        return res.status_code == 200

    def create(self, project_info):
        url = "{}/apps/{}/projects".format(self._api.domain, self._username())
        json = {'name': project_info['name'],
                'access': 'public',
                'profile': {'description': project_info['description']}}
        data, headers = binstar_utils.jencode(json)
        res = self._api.session.post(url, data=data, headers=headers)
        self._check_response(res)
        return res

    def _file_count(self, archive_filename):
        for suffix in (".tar", ".tar.gz", ".tar.bz2"):
            if archive_filename.lower().endswith(suffix):
                with tarfile.open(archive_filename, 'r') as tf:
                    return len(tf.getnames())
        if archive_filename.lower().endswith(".zip"):
            with zipfile.ZipFile(archive_filename, 'r') as zf:
                return len(zf.namelist())
        assert False, ("unsupported archive filename %s" % archive_filename)  # pragma: no cover (should not be reached)

    def stage(self, project_info, archive_filename, uploaded_basename):
        url = "{}/apps/{}/projects/{}/stage".format(self._api.domain, self._username(), project_info['name'])
        config = project_info.copy()
        config['size'] = os.path.getsize(archive_filename)
        file_count = self._file_count(archive_filename)
        if file_count is not None:
            config['num_of_files'] = file_count
        json = {'basename': uploaded_basename, 'configuration': config}
        data, headers = binstar_utils.jencode(json)
        res = self._api.session.post(url, data=data, headers=headers)
        self._check_response(res)
        return res

    def commit(self, project_name, dist_id):

        url = "{}/apps/{}/projects/{}/commit/{}".format(self._api.domain, self._username(), project_name, dist_id)
        data, headers = binstar_utils.jencode({})
        res = self._api.session.post(url, data=data, headers=headers)
        self._check_response(res)
        return res

    def _put_on_s3(self, archive_filename, uploaded_basename, url, s3data):
        with open(archive_filename, 'rb') as f:
            _hexmd5, b64md5, size = binstar_utils.compute_hash(f, size=os.path.getsize(archive_filename))

        s3data = s3data.copy()  # don't modify our parameters
        s3data['Content-Length'] = size
        s3data['Content-MD5'] = b64md5

        with open(archive_filename, 'rb') as archive_file_object:
            data_stream, headers = binstar_requests_ext.stream_multipart(
                s3data,
                files={'file': (uploaded_basename, archive_file_object)})

            res = requests.post(url,
                                data=data_stream,
                                verify=self._api.session.verify,
                                timeout=10 * 60 * 60,
                                headers=headers)
            self._check_response(res)
        return res

    def upload(self, project_info, archive_filename, uploaded_basename):
        """Upload archive_filename created from project, throwing BinstarError."""
        if not self._exists(project_info['name']):
            res = self.create(project_info=project_info)
            assert res.status_code in (200, 201)

        res = self.stage(project_info=project_info,
                         archive_filename=archive_filename,
                         uploaded_basename=uploaded_basename)
        assert res.status_code in (200, 201)

        stage_info = res.json()

        assert 'post_url' in stage_info
        assert 'form_data' in stage_info
        assert 'dist_id' in stage_info

        res = self._put_on_s3(archive_filename,
                              uploaded_basename,
                              url=stage_info['post_url'],
                              s3data=stage_info['form_data'])
        assert res.status_code in (200, 201)

        res = self.commit(project_info['name'], stage_info['dist_id'])
        assert res.status_code in (200, 201)

        return res.json()


class _UploadedStatus(SimpleStatus):
    def __init__(self, json):
        self.url = json.get('url', None)
        logs = []
        if self.url is not None:
            logs.append("Project is at %s" % self.url)
        super(_UploadedStatus, self).__init__(success=True, description="Upload successful.", logs=logs)


# This function is supposed to encapsulate the binstar API (don't
# require any other files to import binstar_client).
# archive_filename is the path to a local tmp file to upload
# uploaded_basename is the filename the server should remember
def _upload(project, archive_filename, uploaded_basename, site=None, username=None, token=None, log_level=None):
    assert not project.problems

    client = _Client(site=site, username=username, token=token, log_level=log_level)
    try:
        json = client.upload(project.publication_info(), archive_filename, uploaded_basename)
        return _UploadedStatus(json)
    except Unauthorized as e:
        return SimpleStatus(success=False,
                            description='Please log in with the "anaconda login" command.',
                            errors=["Not logged in."])
    except BinstarError as e:
        return SimpleStatus(success=False, description="Upload failed.", errors=[str(e)])



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import

import os
import platform
import pytest

import anaconda_project.internal.conda_api as conda_api
import anaconda_project.internal.pip_api as pip_api
from anaconda_project.test.environ_utils import minimal_environ
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents_completing_project_file
from anaconda_project.internal.test.test_conda_api import monkeypatch_conda_not_to_use_links
from anaconda_project.prepare import (prepare_without_interaction, prepare_in_stages, unprepare)
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.project import Project
from anaconda_project import provide
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.providers.conda_env import CondaEnvProvider
from anaconda_project.requirements_registry.requirements.conda_env import CondaEnvRequirement

if platform.system() == 'Windows':
    script_dir = "Scripts"
else:
    script_dir = "bin"

conda_env_var = conda_api.conda_prefix_variable()


def test_find_by_class_name_conda_env():
    registry = RequirementsRegistry()
    found = registry.find_provider_by_class_name(class_name="CondaEnvProvider")
    assert found is not None
    assert isinstance(found, CondaEnvProvider)


@pytest.mark.slow
def test_prepare_and_unprepare_project_scoped_env(monkeypatch):
    monkeypatch_conda_not_to_use_links(monkeypatch)

    def prepare_project_scoped_env(dirname):
        project = Project(dirname)
        fake_old_path = "foo" + os.pathsep + "bar"
        environ = dict(PROJECT_DIR=dirname, PATH=fake_old_path)
        result = prepare_without_interaction(project, environ=environ)
        assert result
        expected_env = os.path.join(dirname, "envs", "default")
        if platform.system() == 'Windows':
            expected_new_path = expected_env + os.pathsep + os.path.join(
                expected_env, script_dir) + os.pathsep + os.path.join(expected_env, "Library",
                                                                      "bin") + os.pathsep + "foo" + os.pathsep + "bar"
        else:
            expected_new_path = os.path.join(expected_env, script_dir) + os.pathsep + "foo" + os.pathsep + "bar"
        expected = dict(PROJECT_DIR=project.directory_path, PATH=expected_new_path)
        conda_api.environ_set_prefix(expected, expected_env)

        expected == result.environ
        assert os.path.exists(os.path.join(expected_env, "conda-meta"))
        conda_meta_mtime = os.path.getmtime(os.path.join(expected_env, "conda-meta"))

        # bare minimum default env shouldn't include these
        # (contrast with the test later where we list them in
        # requirements)
        installed = conda_api.installed(expected_env)
        assert 'ipython' not in installed
        assert 'numpy' not in installed

        # Prepare it again should no-op (use the already-existing environment)
        environ = dict(PROJECT_DIR=dirname, PATH=fake_old_path)
        result = prepare_without_interaction(project, environ=environ)
        assert result
        expected = dict(PROJECT_DIR=project.directory_path, PATH=expected_new_path)
        conda_api.environ_set_prefix(expected, expected_env)
        assert expected == result.environ
        assert conda_meta_mtime == os.path.getmtime(os.path.join(expected_env, "conda-meta"))

        # Now unprepare
        status = unprepare(project, result)
        assert status
        assert status.status_description == ('Deleted environment files in %s.' % (expected_env))
        assert status.errors == []
        assert not os.path.exists(expected_env)

    with_directory_contents_completing_project_file(dict(), prepare_project_scoped_env)


def test_prepare_project_scoped_env_conda_create_fails(monkeypatch):
    def mock_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        raise conda_api.CondaError("error_from_conda_create")

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_create)

    def prepare_project_scoped_env_fails(dirname):
        project = Project(dirname)
        environ = minimal_environ(PROJECT_DIR=dirname)
        result = prepare_without_interaction(project, environ=environ)
        assert not result

        assert 'CONDA_DEFAULT_ENV' not in result.environ
        assert 'CONDA_ENV_PATH' not in result.environ

        # unprepare should not have anything to do
        status = unprepare(project, result)
        assert status
        assert status.errors == []
        assert status.status_description == "Nothing to clean up for environment 'default'."

    with_directory_contents_completing_project_file(dict(), prepare_project_scoped_env_fails)


def test_unprepare_gets_error_on_delete(monkeypatch):
    def mock_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        os.makedirs(os.path.join(prefix, "conda-meta"))

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_create)

    def prepare_project_scoped_env(dirname):
        project = Project(dirname)
        environ = minimal_environ(PROJECT_DIR=dirname)
        result = prepare_without_interaction(project, environ=environ)
        assert result
        expected_env = os.path.join(dirname, "envs", "default")

        # Now unprepare

        def mock_rmtree(path):
            raise IOError("I will never rm the tree!")

        monkeypatch.setattr('shutil.rmtree', mock_rmtree)

        status = unprepare(project, result)
        assert status.status_description == ('Failed to remove environment files in %s: I will never rm the tree!.' %
                                             (expected_env))
        assert not status

        assert os.path.exists(expected_env)

        # so we can rmtree our tmp directory
        monkeypatch.undo()

    with_directory_contents_completing_project_file(dict(), prepare_project_scoped_env)


def test_prepare_project_scoped_env_not_attempted_in_check_mode(monkeypatch):
    def mock_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        raise Exception("Should not have attempted to create env")

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_create)

    def prepare_project_scoped_env_not_attempted(dirname):
        project = Project(dirname)
        environ = minimal_environ(PROJECT_DIR=dirname)
        result = prepare_without_interaction(project, environ=environ, mode=provide.PROVIDE_MODE_CHECK)
        assert not result
        expected_env_path = os.path.join(dirname, "envs", "default")
        assert [
            ('missing requirement to run this project: ' +
             'The project needs a Conda environment containing all required packages.'),
            "  '%s' doesn't look like it contains a Conda environment yet." % expected_env_path
        ] == result.errors

        # unprepare should not have anything to do
        status = unprepare(project, result)
        assert status
        assert status.errors == []
        assert status.status_description == ("Nothing to clean up for environment 'default'.")

    with_directory_contents_completing_project_file(dict(), prepare_project_scoped_env_not_attempted)


@pytest.mark.slow
def test_prepare_project_scoped_env_with_packages(monkeypatch):
    monkeypatch_conda_not_to_use_links(monkeypatch)

    def prepare_project_scoped_env_with_packages(dirname):
        project = Project(dirname)
        environ = minimal_environ(PROJECT_DIR=dirname)
        result = prepare_without_interaction(project, environ=environ)
        assert result

        prefix = result.environ[conda_env_var]
        installed = conda_api.installed(prefix)

        assert 'ipython' in installed
        assert 'numpy' in installed
        assert 'bokeh' not in installed

        # Preparing it again with new packages added should add those
        deps = project.project_file.get_value('packages')
        project.project_file.set_value('packages', deps + ['bokeh'])
        project.project_file.save()
        environ = minimal_environ(PROJECT_DIR=dirname)
        result = prepare_without_interaction(project, environ=environ)
        assert result

        prefix = result.environ[conda_env_var]
        installed = conda_api.installed(prefix)

        assert 'ipython' in installed
        assert 'numpy' in installed
        assert 'bokeh' in installed

        installed_pip = pip_api.installed(prefix)
        assert 'flake8' in installed_pip

        # Preparing it again with a bogus package should fail
        deps = project.project_file.get_value('packages')
        project.project_file.set_value(['packages'], deps + ['boguspackage'])
        project.project_file.save()
        environ = minimal_environ(PROJECT_DIR=dirname)
        result = prepare_without_interaction(project, environ=environ)
        assert not result

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
packages:
    - ipython
    - numpy
    - pip:
      - flake8
"""}, prepare_project_scoped_env_with_packages)


def _conda_env_status(prepare_context):
    for status in prepare_context.statuses:
        if isinstance(status.requirement, CondaEnvRequirement):
            return status
    raise AssertionError("no CondaEnvRequirement found")


def test_configure_inherited(monkeypatch):
    def mock_conda_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        from anaconda_project.internal.makedirs import makedirs_ok_if_exists
        metadir = os.path.join(prefix, "conda-meta")
        makedirs_ok_if_exists(metadir)
        for p in pkgs:
            pkgmeta = os.path.join(metadir, "%s-0.1-pyNN.json" % p)
            open(pkgmeta, 'a').close()

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_conda_create)

    def check(dirname):
        envs_dir = os.path.join(dirname, "envs")

        project = Project(dirname)
        environ = minimal_environ(PROJECT_DIR=dirname)
        stage = prepare_in_stages(project, environ=environ)

        prepare_context = stage.configure()

        status = _conda_env_status(prepare_context)
        req = status.requirement
        provider = status.provider

        # check initial config

        config = provider.read_config(req, prepare_context.environ, prepare_context.local_state_file,
                                      prepare_context.default_env_spec_name, prepare_context.overrides)

        assert dict(env_name='default', source='project', value=os.path.join(envs_dir, 'default')) == config

        # set inherited mode

        config['source'] = 'inherited'

        provider.set_config_values_as_strings(req, prepare_context.environ, prepare_context.local_state_file,
                                              prepare_context.default_env_spec_name, prepare_context.overrides, config)

        config = provider.read_config(req, prepare_context.environ, prepare_context.local_state_file,
                                      prepare_context.default_env_spec_name, prepare_context.overrides)

        assert dict(env_name='default', source='inherited', value=os.environ.get(req.env_var)) == config

        # disable inherited mode again

        config['source'] = 'project'
        config['env_name'] = 'default'

        provider.set_config_values_as_strings(req, prepare_context.environ, prepare_context.local_state_file,
                                              prepare_context.default_env_spec_name, prepare_context.overrides, config)

        config = provider.read_config(req, prepare_context.environ, prepare_context.local_state_file,
                                      prepare_context.default_env_spec_name, prepare_context.overrides)

        assert dict(env_name='default', source='project', value=os.path.join(envs_dir, 'default')) == config

    with_directory_contents_completing_project_file(dict(), check)


def test_configure_different_env_spec(monkeypatch):
    def mock_conda_create(prefix, pkgs, channels, stdout_callback, stderr_callback):
        from anaconda_project.internal.makedirs import makedirs_ok_if_exists
        metadir = os.path.join(prefix, "conda-meta")
        makedirs_ok_if_exists(metadir)
        for p in pkgs:
            pkgmeta = os.path.join(metadir, "%s-0.1-pyNN.json" % p)
            open(pkgmeta, 'a').close()

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_conda_create)

    def check(dirname):
        envs_dir = os.path.join(dirname, "envs")

        project = Project(dirname)
        environ = minimal_environ(PROJECT_DIR=dirname)
        stage = prepare_in_stages(project, environ=environ)

        prepare_context = stage.configure()

        status = _conda_env_status(prepare_context)
        req = status.requirement
        provider = status.provider

        config = provider.read_config(req, prepare_context.environ, prepare_context.local_state_file,
                                      prepare_context.default_env_spec_name, prepare_context.overrides)
        assert dict(env_name='default', source='project', value=os.path.join(envs_dir, 'default')) == config

        config['env_name'] = 'bar'

        provider.set_config_values_as_strings(req, prepare_context.environ, prepare_context.local_state_file,
                                              prepare_context.default_env_spec_name, prepare_context.overrides, config)

        config = provider.read_config(req, prepare_context.environ, prepare_context.local_state_file,
                                      prepare_context.default_env_spec_name, prepare_context.overrides)
        assert dict(env_name='bar', source='project', value=os.path.join(envs_dir, 'bar')) == config

        assert os.path.join(envs_dir, 'bar') == prepare_context.local_state_file.get_value(['variables', req.env_var])

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
env_specs:
  default:
    packages: []
    channels: []
  foo:
    packages: []
    channels: []
  bar:
    packages: []
    channels: []
"""
        }, check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Internal file and directory utilities."""
from __future__ import absolute_import, print_function

import os
import platform


def directory_contains_subdirectory(parent, child):
    """Test whether child is somewhere underneath parent."""
    parent = os.path.realpath(parent)
    child = os.path.realpath(child)

    # note: there's an os.path.commonprefix() but it's useless
    # because it's character-based so it thinks /foo is the common
    # prefix of /foo and /foobar.
    def _helper(real_parent, real_child):
        dirname = os.path.dirname(real_child)
        if dirname == real_parent:
            return True
        elif len(dirname) < len(real_parent):
            return False
        else:
            return _helper(real_parent, dirname)

    return _helper(parent, child)


def subdirectory_relative_to_directory(child, parent):
    """Make subdirectory name relative to the given parent."""
    parent = os.path.realpath(parent)
    child = os.path.realpath(child)

    if not directory_contains_subdirectory(parent, child):
        return child

    assert child.startswith(parent)

    child = child[len(parent):]
    if child.startswith("/"):  # on both unix and windows
        child = child[1:]
    if platform.system() == 'Windows' and child.startswith("\\"):
        child = child[1:]  # pragma: no cover (too hard to test on linux)
    return child



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Frontend class representing a UX."""
from __future__ import absolute_import

from abc import ABCMeta, abstractmethod

from anaconda_project.internal.metaclass import with_metaclass


class Frontend(with_metaclass(ABCMeta)):
    """A UX (CLI, GUI, etc.) for project operations."""

    def __init__(self):
        """Construct a Frontend."""
        self._info_buf = ''
        self._error_buf = ''

    def _partial(self, data, buf, line_handler):
        buf = buf + data
        (start, sep, end) = buf.partition('\n')
        while sep != '':
            # we do this instead of using os.linesep in case
            # something on windows outputs unix-style line
            # endings, we don't want to go haywire.  On unix when
            # we actually want \r to carriage return, we'll be
            # overriding this "partial" handler and not using this
            # buffering implementation.
            if start.endswith('\r'):
                start = start[:-1]
            line_handler(start)
            buf = end
            (start, sep, end) = buf.partition('\n')
        return buf

    def partial_info(self, data):
        """Log only part of an info-level line.

        The default implementation buffers this until a line separator
        and then passes the entire line to info().
        Subtypes can override this if they want to print output
        immediately as it arrives.
        """
        self._info_buf = self._partial(data, self._info_buf, self.info)

    def partial_error(self, data):
        """Log only part of an error-level line.

        The default implementation buffers this until a line separator
        and then passes the entire line to error().
        Subtypes can override this if they want to print output
        immediately as it arrives.
        """
        self._error_buf = self._partial(data, self._error_buf, self.error)

    @abstractmethod
    def info(self, message):
        """Log an info-level message."""
        pass  # pragma: no cover

    @abstractmethod
    def error(self, message):
        """Log an error-level message.

        A rule of thumb is that if a function also returns a
        ``Status``, this message should also be appended to the
        ``errors`` field on that status.
        """
        pass  # pragma: no cover

    # @abstractmethod
    # def new_progress(self):
    #    """Create an appropriate subtype of Progress."""
    #    pass  # pragma: no cover


class NullFrontend(Frontend):
    """A frontend that doesn't do anything."""

    def __init__(self):
        """Construct a null frontend."""
        super(NullFrontend, self).__init__()

    def partial_info(self, data):
        """Part of a log message."""
        pass

    def partial_error(self, data):
        """Part of an error message."""
        pass

    def info(self, message):
        """Log an info-level message."""
        pass

    def error(self, message):
        """Log an error-level message."""
        pass


_singleton_null_frontend = None


def _null_frontend():
    global _singleton_null_frontend
    if _singleton_null_frontend is None:
        _singleton_null_frontend = NullFrontend()
    return _singleton_null_frontend


class _ErrorRecordingFrontendProxy(Frontend):
    def __init__(self, underlying):
        super(_ErrorRecordingFrontendProxy, self).__init__()
        self._errors = []
        self.underlying = underlying

    def info(self, message):
        """Log an info-level message."""
        self.underlying.info(message)

    def error(self, message):
        """Log an error-level message."""
        self._errors.append(message)
        self.underlying.error(message)

    def pop_errors(self):
        result = self._errors
        self._errors = []
        return result


def _new_error_recorder(frontend):
    return _ErrorRecordingFrontendProxy(frontend)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from anaconda_project.local_state_file import LocalStateFile
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirement import UserConfigOverrides
from anaconda_project.requirements_registry.requirements.redis import RedisRequirement

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents


def test_find_by_service_type_redis():
    registry = RequirementsRegistry()
    found = registry.find_requirement_by_service_type(service_type='redis', env_var='MYREDIS', options=dict())
    assert found is not None
    assert isinstance(found, RedisRequirement)
    assert found.env_var == 'MYREDIS'
    assert found.service_type == 'redis'


def test_redis_url_not_set():
    def check_not_set(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = RedisRequirement(registry=RequirementsRegistry(), env_var="REDIS_URL")
        status = requirement.check_status(dict(), local_state, 'default', UserConfigOverrides())
        assert not status
        assert "Environment variable REDIS_URL is not set." == status.status_description

    with_directory_contents({}, check_not_set)


def test_redis_url_bad_scheme():
    def check_bad_scheme(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = RedisRequirement(registry=RequirementsRegistry(), env_var="REDIS_URL")
        status = requirement.check_status(
            dict(REDIS_URL="http://example.com/"),
            local_state,
            'default',
            UserConfigOverrides())
        assert not status
        assert "REDIS_URL value 'http://example.com/' does not have 'redis:' scheme." == status.status_description

    with_directory_contents({}, check_bad_scheme)


def _monkeypatch_can_connect_to_socket_fails(monkeypatch):
    can_connect_args_list = []

    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        can_connect_args = dict(host=host, port=port, timeout_seconds=timeout_seconds)
        can_connect_args_list.append(can_connect_args)
        return False

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)

    return can_connect_args_list


def test_redis_url_cannot_connect(monkeypatch):
    def check_cannot_connect(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = RedisRequirement(registry=RequirementsRegistry(), env_var="REDIS_URL")
        can_connect_args_list = _monkeypatch_can_connect_to_socket_fails(monkeypatch)
        status = requirement.check_status(
            dict(REDIS_URL="redis://example.com:1234/"),
            local_state,
            'default',
            UserConfigOverrides())
        assert dict(host='example.com', port=1234, timeout_seconds=0.5) == can_connect_args_list[0]
        assert dict(host='localhost', port=6379, timeout_seconds=0.5) == can_connect_args_list[1]

        assert not status
        expected = "Cannot connect to Redis at redis://example.com:1234/."
        assert expected == status.status_description

    with_directory_contents({}, check_cannot_connect)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from anaconda_project.project import Project
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.requirements_registry.requirements.service import ServiceRequirement

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents_completing_project_file


def _load_service_requirement(dirname):
    project = Project(dirname)
    assert [] == project.problems
    for req in project.requirements(project.default_env_spec_name):
        if isinstance(req, ServiceRequirement):
            return req
    raise RuntimeError("no ServiceRequirement found")


def test_service_dict_with_options():
    def check(dirname):
        req = _load_service_requirement(dirname)
        assert req.options == dict(type='redis', foo='bar', default='hello')
        assert req.env_var == 'FOOBAR'

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
    FOOBAR: { type: redis, foo: bar, default: hello }
    """}, check)


def test_service_dict_with_bad_value():
    def check(dirname):
        project = Project(dirname)
        assert ["Service FOOBAR should have a service type string or a dictionary as its value."] == project.problems

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
    FOOBAR:
     - 42
    """}, check)


def test_service_with_bad_service_type():
    def check(dirname):
        project = Project(dirname)
        assert ["Service FOOBAR has an unknown type 'not_a_service'."] == project.problems

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
    FOOBAR: not_a_service
    """}, check)


def test_service_dict_with_no_service_type():
    def check(dirname):
        project = Project(dirname)
        assert ["Service FOOBAR doesn't contain a 'type' field."] == project.problems

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
    FOOBAR: {}
    """}, check)


def test_service_dict_bad_default():
    def check(dirname):
        project = Project(dirname)
        assert ["default value for variable FOOBAR must be null, a string, or a number, not []."] == project.problems

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
    FOOBAR: { type: redis, default: [] }
    """}, check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os

from anaconda_project.test.project_utils import project_dir_disable_dedicated_env
from anaconda_project.test.environ_utils import (minimal_environ, minimal_environ_no_conda_env)
from anaconda_project.env_spec import EnvSpec
from anaconda_project.local_state_file import LocalStateFile
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirement import UserConfigOverrides
from anaconda_project.requirements_registry.requirements.conda_env import CondaEnvRequirement

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents
from anaconda_project.internal import conda_api

conda_env_var = conda_api.conda_prefix_variable()


def _empty_default_requirement():
    return CondaEnvRequirement(registry=RequirementsRegistry(), env_specs=dict(default=EnvSpec('default', [], [])))


def test_env_var():
    registry = RequirementsRegistry()
    requirement = CondaEnvRequirement(registry)
    assert requirement.env_var == conda_env_var


def test_conda_env_title_and_description():
    requirement = _empty_default_requirement()
    assert requirement.title == 'A Conda environment'
    assert requirement.description == 'The project needs a Conda environment containing all required packages.'


def test_conda_default_env_not_set():
    def check_conda_default_env_not_set(dirname):
        requirement = _empty_default_requirement()
        project_dir_disable_dedicated_env(dirname)
        local_state = LocalStateFile.load_for_directory(dirname)
        status = requirement.check_status(
            minimal_environ_no_conda_env(PROJECT_DIR=dirname),
            local_state,
            'default',
            UserConfigOverrides())
        expected = "'{}' doesn't look like it contains a Conda environment yet.".format(os.path.join(dirname, 'envs',
                                                                                                     'default'))
        assert expected == status.status_description

    with_directory_contents(dict(), check_conda_default_env_not_set)


def test_conda_default_env_is_bogus():
    def check_conda_default_env_is_bogus(dirname):
        requirement = _empty_default_requirement()
        project_dir_disable_dedicated_env(dirname)
        local_state = LocalStateFile.load_for_directory(dirname)
        status = requirement.check_status(
            minimal_environ_no_conda_env(**{'PROJECT_DIR': dirname}),
            local_state,
            'default',
            UserConfigOverrides(inherited_env="not_a_real_env_anyone_has"))
        expected = "'not_a_real_env_anyone_has' doesn't look like it contains a Conda environment yet."
        assert expected == status.status_description

    with_directory_contents(dict(), check_conda_default_env_is_bogus)


def test_conda_fails_while_listing_installed(monkeypatch):
    def check_fails_while_listing_installed(dirname):
        def sabotaged_installed_command(prefix):
            from anaconda_project.internal import conda_api
            raise conda_api.CondaError("sabotage!")

        monkeypatch.setattr('anaconda_project.internal.conda_api.installed', sabotaged_installed_command)

        project_dir_disable_dedicated_env(dirname)
        local_state = LocalStateFile.load_for_directory(dirname)

        requirement = CondaEnvRequirement(registry=RequirementsRegistry(),
                                          env_specs=dict(default=EnvSpec('default', ['not_a_real_package'], [])))
        environ = minimal_environ(PROJECT_DIR=dirname)
        status = requirement.check_status(environ,
                                          local_state,
                                          'default',
                                          UserConfigOverrides(inherited_env=environ.get(conda_env_var)))
        assert status.status_description.startswith("Conda failed while listing installed packages in ")
        assert status.status_description.endswith(": sabotage!")

    with_directory_contents(dict(), check_fails_while_listing_installed)


def test_missing_package():
    def check_missing_package(dirname):
        requirement = CondaEnvRequirement(
            registry=RequirementsRegistry(),
            env_specs=dict(default=EnvSpec('default', ['boguspackage', 'boguspackage2'], [])))
        project_dir_disable_dedicated_env(dirname)
        local_state = LocalStateFile.load_for_directory(dirname)
        environ = minimal_environ(PROJECT_DIR=dirname)
        status = requirement.check_status(environ,
                                          local_state,
                                          'default',
                                          UserConfigOverrides(inherited_env=environ.get(conda_env_var)))
        assert "Conda environment is missing packages: boguspackage, boguspackage2" == status.status_description

    with_directory_contents(dict(), check_missing_package)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Network utilities for use by plugins."""
import socket


def _get_urlparse():
    try:
        # py3
        import urllib.parse
        return urllib.parse  # pragma: no cover
    except ImportError:  # pragma: no cover
        # py2
        import urlparse
        return urlparse


urlparse = _get_urlparse()


def can_connect_to_socket(host, port, timeout_seconds=0.5):
    """Check whether we can connect to a server at host:port.

    Args:
        host (str): the host
        port (int): the port
        timeout_seconds (float): how long to wait for failure
    Returns:
        True if we could connect
    """
    try:
        s = socket.create_connection(address=(host, port), timeout=timeout_seconds)
        s.close()
        return True
    except IOError:
        return False



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import pytest

from anaconda_project.internal.toposort import toposort_from_dependency_info, CycleError


# sort tuples of the form (thing, (dep1, dep2))
def sort_tuples(tuples, can_ignore=None):
    def get_node_key(t):
        return t[0]

    def get_dependency_keys(t):
        return t[1]

    def can_ignore_key(k):
        return k in can_ignore

    if can_ignore is None:
        can_ignore_func = None
    else:
        can_ignore_func = can_ignore_key

    return list(map(lambda t: t[0], toposort_from_dependency_info(tuples, get_node_key, get_dependency_keys,
                                                                  can_ignore_func)))


def test_empty():
    sorted = sort_tuples([])
    assert [] == sorted


def test_one():
    unsorted = [(1, ())]
    sorted = sort_tuples(unsorted)
    assert [1] == sorted


def test_two():
    unsorted = [(1, (2, )), (2, ())]
    sorted = sort_tuples(unsorted)
    assert [2, 1] == sorted


def test_three():
    unsorted = [(1, (2, 3)), (2, (3, )), (3, ())]
    sorted = sort_tuples(unsorted)
    assert [3, 2, 1] == sorted


def test_four():
    unsorted = [(1, (2, 3)), (2, (3, 4)), (3, ()), (4, ())]
    sorted = sort_tuples(unsorted)
    assert [4, 3, 2, 1] == sorted or [3, 4, 2, 1] == sorted


def test_cycle():
    unsorted = [(1, (2, )), (2, (1, ))]
    with pytest.raises(CycleError) as excinfo:
        sort_tuples(unsorted)
    assert 'Cycle in graph' in repr(excinfo.value)
    assert isinstance(excinfo.value.involving, tuple)


def test_duplicate_nodes():
    unsorted = [(1, (2, )), (1, ()), (2, ())]
    with pytest.raises(ValueError) as excinfo:
        sort_tuples(unsorted)
    assert 'two nodes with the same key 1' in repr(excinfo.value)


def test_dependency_not_in_list():
    unsorted = [(1, (2, ))]
    with pytest.raises(ValueError) as excinfo:
        sort_tuples(unsorted)
    assert 'Dependency 2 was not in the list of nodes' in repr(excinfo.value)


def test_dependency_not_in_list_but_can_ignore_that():
    unsorted = [(1, (2, ))]
    sorted = sort_tuples(unsorted, can_ignore=set([2]))
    assert [1] == sorted



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import sys

from anaconda_project.project import Project
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.internal.cli.project_load import load_project

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents


def _monkeypatch_input(monkeypatch, answer):
    answers = []
    if answer is None or isinstance(answer, str):
        answers.append(answer)
    else:
        answers = list(answer)
        answers.reverse()

    def mock_input(prompt):
        sys.stdout.write(prompt)
        item = answers.pop()
        if item is None:
            raise EOFError("eof on input")
        else:
            return item

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils._input', mock_input)


def test_interactively_fix_project(monkeypatch, capsys):
    def check(dirname):

        broken_project = Project(dirname)
        assert len(broken_project.fixable_problems) == 1

        def mock_isatty_true():
            return True

        # python 2 can throw a "readonly" error if you try to patch sys.stdin.isatty itself
        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_isatty_true)
        _monkeypatch_input(monkeypatch, ["y"])

        project = load_project(dirname)
        assert project.problems == []

        out, err = capsys.readouterr()
        assert out == ("%s: The env_specs section is empty.\nAdd an environment spec to anaconda-project.yml? " %
                       DEFAULT_PROJECT_FILENAME)
        assert err == ""

    with_directory_contents({DEFAULT_PROJECT_FILENAME: "name: foo"}, check)


def test_interactively_no_fix_project(monkeypatch, capsys):
    def check(dirname):

        broken_project = Project(dirname)
        assert len(broken_project.fixable_problems) == 1

        def mock_isatty_true():
            return True

        # python 2 can throw a "readonly" error if you try to patch sys.stdin.isatty itself
        monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_isatty_true)
        _monkeypatch_input(monkeypatch, ["n"])

        project = load_project(dirname)
        assert project.problems == ["%s: The env_specs section is empty." % DEFAULT_PROJECT_FILENAME]

        out, err = capsys.readouterr()
        assert out == ("%s: The env_specs section is empty.\nAdd an environment spec to anaconda-project.yml? " %
                       DEFAULT_PROJECT_FILENAME)
        assert err == ""

    with_directory_contents({DEFAULT_PROJECT_FILENAME: "name: foo\nplatforms: [linux-64,osx-64,win-64]\n"}, check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Library for working with projects."""

_beta_test_mode = False


def _enter_beta_test_mode():
    """Called by anaconda-project executable to do special things for beta."""
    global _beta_test_mode
    _beta_test_mode = True



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""A requirement for a service."""

from copy import deepcopy
from anaconda_project.requirements_registry.requirement import EnvVarRequirement

from anaconda_project.internal.py2_compat import is_string


class ServiceRequirement(EnvVarRequirement):
    """Abstract base class for a requirement from the services section of the project file."""

    @classmethod
    def _parse(cls, varname, item, problems):
        """Parse an item from the services: section."""
        service_type = None
        if is_string(item):
            service_type = item
            options = dict(type=service_type)
        elif isinstance(item, dict):
            service_type = item.get('type', None)
            if service_type is None:
                problems.append("Service {} doesn't contain a 'type' field.".format(varname))
                return None
            options = deepcopy(item)
        else:
            problems.append("Service {} should have a service type string or a dictionary as its value.".format(
                varname))
            return None

        if not EnvVarRequirement._parse_default(options, varname, problems):
            return None

        return dict(service_type=service_type, env_var=varname, options=options)

    @property
    def service_type(self):
        """Get service type string."""
        return self.options['type']

    @property
    def ignore_patterns(self):
        """Override superclass with our ignore patterns."""
        return set(['/services/'])



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os

from anaconda_project import archiver
from anaconda_project import project_ops
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents
from anaconda_project.internal.test.fake_frontend import FakeFrontend


def test_parse_ignore_file():
    def check(dirname):
        frontend = FakeFrontend()
        patterns = archiver._parse_ignore_file(os.path.join(dirname, ".projectignore"), frontend)
        assert [] == frontend.errors

        pattern_strings = [pattern.pattern for pattern in patterns]

        assert pattern_strings == ['bar', '/baz', 'whitespace_surrounding',
                                   'foo # this comment will be part of the pattern', '#patternwithhash', 'hello']

    with_directory_contents(
        {".projectignore": """
# this is a sample .projectignore
   # there can be whitespace before the comment
bar
/baz
   whitespace_surrounding%s
foo # this comment will be part of the pattern
\#patternwithhash

# blank line above me

hello

        """ % ("   ")}, check)


def test_parse_missing_ignore_file():
    def check(dirname):
        frontend = FakeFrontend()
        patterns = archiver._parse_ignore_file(os.path.join(dirname, ".projectignore"), frontend)
        assert [] == frontend.errors

        pattern_strings = [pattern.pattern for pattern in patterns]

        assert pattern_strings == []

    with_directory_contents(dict(), check)


def test_parse_ignore_file_with_io_error(monkeypatch):
    def check(dirname):
        frontend = FakeFrontend()
        ignorefile = os.path.join(dirname, ".projectignore")

        from codecs import open as real_open

        def mock_codecs_open(*args, **kwargs):
            if args[0].endswith(".projectignore"):
                raise IOError("NOPE")
            else:
                return real_open(*args, **kwargs)

        monkeypatch.setattr('codecs.open', mock_codecs_open)

        patterns = archiver._parse_ignore_file(ignorefile, frontend)
        assert patterns is None
        assert ["Failed to read %s: NOPE" % ignorefile] == frontend.errors

        # enable cleaning it up
        os.chmod(ignorefile, 0o777)

    with_directory_contents({".projectignore": ""}, check)


def test_parse_default_ignore_file():
    def check(dirname):
        project_ops._add_projectignore_if_none(dirname)
        ignorefile = os.path.join(dirname, ".projectignore")
        assert os.path.isfile(ignorefile)

        frontend = FakeFrontend()
        patterns = archiver._parse_ignore_file(ignorefile, frontend)
        assert [] == frontend.errors

        pattern_strings = [pattern.pattern for pattern in patterns]

        assert pattern_strings == ['/anaconda-project-local.yml', '__pycache__/', '*.pyc', '*.pyo', '*.pyd',
                                   '/.ipynb_checkpoints', '/.spyderproject']

    with_directory_contents(dict(), check)


def _test_file_pattern_matcher(tests, is_directory):
    class FakeInfo(object):
        pass

    for pattern_string in tests.keys():
        pattern = archiver._FilePattern(pattern_string)
        should_match = tests[pattern_string]['yes']
        should_not_match = tests[pattern_string]['no']
        matched = []
        did_not_match = []
        for filename in (should_match + should_not_match):
            info = FakeInfo()
            setattr(info, 'unixified_relative_path', filename)
            setattr(info, 'is_directory', is_directory)
            if pattern.matches(info):
                matched.append(filename)
            else:
                did_not_match.append(filename)
        assert should_match == matched
        assert should_not_match == did_not_match


def test_file_pattern_matcher_non_directories():
    tests = {
        'foo': {
            'yes': ['foo', 'bar/foo', 'foo/bar'],
            'no': ['bar', 'foobar', 'barfoo']
        },
        '/foo': {
            'yes': ['foo', 'foo/bar'],
            'no': ['barfoo', 'bar/foo', 'bar', 'foobar']
        },
        'foo/': {
            'yes': [],
            'no': ['foo', 'barfoo', 'bar/foo', 'foo/bar', 'bar', 'foobar']
        },
        '/foo/': {
            'yes': [],
            'no': ['foo', 'barfoo', 'bar/foo', 'foo/bar', 'bar', 'foobar']
        },
    }

    _test_file_pattern_matcher(tests, is_directory=False)


def test_file_pattern_matcher_with_directories():
    tests = {
        'foo': {
            'yes': ['foo', 'bar/foo', 'foo/bar'],
            'no': ['bar', 'foobar', 'barfoo']
        },
        '/foo': {
            'yes': ['foo', 'foo/bar'],
            'no': ['barfoo', 'bar/foo', 'bar', 'foobar']
        }
    }

    # we'll say these are all dirs, so trailing / shouldn't matter
    tests['foo/'] = tests['foo']
    tests['/foo/'] = tests['/foo']

    _test_file_pattern_matcher(tests, is_directory=True)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import sys

import anaconda_project.internal.cli.console_utils as console_utils


def test_stdin_is_interactive(monkeypatch):
    result = console_utils.stdin_is_interactive()
    assert result is True or result is False


def _monkeypatch_input(monkeypatch, answer):
    answers = []
    if answer is None or isinstance(answer, str):
        answers.append(answer)
    else:
        answers = list(answer)
        answers.reverse()

    def mock_input(prompt):
        sys.stdout.write(prompt)
        item = answers.pop()
        if item is None:
            raise EOFError("eof on input")
        else:
            return item

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils._input', mock_input)


def _monkeypatch_getpass(monkeypatch, answer):
    answers = []
    if answer is None or isinstance(answer, str):
        answers.append(answer)
    else:
        answers = list(answer)
        answers.reverse()

    def mock_getpass(prompt, stream=None):
        sys.stderr.write(prompt)
        item = answers.pop()
        if item is None:
            raise EOFError("eof on getpass")
        else:
            return item

    monkeypatch.setattr('getpass.getpass', mock_getpass)


def test_console_yes_or_no(monkeypatch, capsys):
    def mock_isatty_true():
        return True

    # python 2 can throw a "readonly" error if you try to patch sys.stdin.isatty itself
    monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_isatty_true)

    for answer in ("y", "Y", "yes", "Yes", "YES", "yoyo"):
        _monkeypatch_input(monkeypatch, answer)
        assert console_utils.console_ask_yes_or_no("foo?", False)
        out, err = capsys.readouterr()
        assert out == "foo? "
        assert err == ""

    for answer in ("n", "N", "no", "No", "NO", "nay"):
        _monkeypatch_input(monkeypatch, answer)
        assert not console_utils.console_ask_yes_or_no("foo?", True)
        out, err = capsys.readouterr()
        assert out == "foo? "
        assert err == ""

    _monkeypatch_input(monkeypatch, ("", "yes"))
    assert console_utils.console_ask_yes_or_no("foo?", False)
    out, err = capsys.readouterr()
    assert out == "foo? foo? (enter y or n): "
    assert err == ""

    _monkeypatch_input(monkeypatch, None)
    assert console_utils.console_ask_yes_or_no("foo?", True)
    out, err = capsys.readouterr()
    assert out == "foo? "
    assert err == ""

    _monkeypatch_input(monkeypatch, None)
    assert not console_utils.console_ask_yes_or_no("foo?", False)
    out, err = capsys.readouterr()
    assert out == "foo? "
    assert err == ""


def test_format_names_and_descriptions():
    class Thing(object):
        def __init__(self, name, description):
            self.name = name
            self.description = description

    cases = [
        ([], "\n"), ([Thing("a", "b")], "Name  Description\n====  ===========\na     b\n"),
        ([Thing("cd", "ef"), Thing("a", "b")], "Name  Description\n====  ===========\na     b\ncd    ef\n")
    ]
    for case in cases:
        assert console_utils.format_names_and_descriptions(case[0]) == case[1]


def test_console_get_password(monkeypatch, capsys):
    def mock_isatty_true():
        return True

    # python 2 can throw a "readonly" error if you try to patch sys.stdin.isatty itself
    monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_isatty_true)

    _monkeypatch_getpass(monkeypatch, "s3kr3t")

    password = console_utils.console_input("foo: ", encrypted=True)

    assert password == 's3kr3t'

    (out, err) = capsys.readouterr()

    assert out == ''
    assert err == 'foo: '



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from anaconda_project.project import Project
from anaconda_project.local_state_file import LocalStateFile
from anaconda_project.internal.test.fake_frontend import FakeFrontend


def project_dir_disable_dedicated_env(dirname):
    """Modify project config to disable having a dedicated environment."""
    local_state = LocalStateFile.load_for_directory(dirname)
    local_state.set_value('inherit_environment', True)
    local_state.save()


def project_no_dedicated_env(*args, **kwargs):
    """Get a project that won't create envs/default as long as there's an env already."""
    if len(args) > 0:
        dirname = args[0]
    elif 'directory_path' in kwargs:
        dirname = kwargs['directory_path']
    else:
        raise RuntimeError("no directory_path for Project")

    project_dir_disable_dedicated_env(dirname)

    if 'frontend' not in kwargs:
        kwargs['frontend'] = FakeFrontend()

    project = Project(*args, **kwargs)

    return project



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import codecs
import os
import platform
import sys

from anaconda_project.test.project_utils import project_no_dedicated_env
from anaconda_project.internal.test.tmpfile_utils import (with_directory_contents,
                                                          with_directory_contents_completing_project_file)
from anaconda_project.test.environ_utils import minimal_environ, strip_environ
from anaconda_project.local_state_file import DEFAULT_LOCAL_STATE_FILENAME
from anaconda_project.local_state_file import LocalStateFile
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirement import UserConfigOverrides
from anaconda_project.requirements_registry.providers.redis import RedisProvider
from anaconda_project.requirements_registry.requirements.redis import RedisRequirement
from anaconda_project.prepare import prepare_without_interaction, unprepare
from anaconda_project import provide
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME


# This is kind of an awkward way to do it for historical reasons,
# we print out the logs/errors captured by FakeFrontend, instead
# of rewriting the tests in here to have a frontend that prints.
def _prepare_printing_errors(project, environ=None, mode=provide.PROVIDE_MODE_DEVELOPMENT):
    result = prepare_without_interaction(project, environ=environ, mode=mode)
    for message in project.frontend.logs:
        print(message)
    for error in project.frontend.errors:
        print(error, file=sys.stderr)
    if not result:
        assert result.errors == project.frontend.errors
    return result


def _redis_requirement():
    return RedisRequirement(registry=RequirementsRegistry(), env_var="REDIS_URL")


def test_reading_default_config():
    def read_config(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = _redis_requirement()
        provider = RedisProvider()
        config = provider.read_config(requirement, dict(), local_state, 'default', UserConfigOverrides())
        assert 6380 == config['lower_port']
        assert 6449 == config['upper_port']

    with_directory_contents(dict(), read_config)


def test_reading_valid_config():
    def read_config(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = _redis_requirement()
        provider = RedisProvider()
        config = provider.read_config(requirement, dict(), local_state, 'default', UserConfigOverrides())
        assert 7389 == config['lower_port']
        assert 7421 == config['upper_port']
        assert 'find_all' == config['source']

    with_directory_contents(
        {
            DEFAULT_LOCAL_STATE_FILENAME: """
service_options:
  REDIS_URL:
    port_range: 7389-7421
    autostart: false
         """
        }, read_config)


def _read_invalid_port_range(capsys, port_range):
    def read_config(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = _redis_requirement()
        provider = RedisProvider()
        config = provider.read_config(requirement, dict(), local_state, 'default', UserConfigOverrides())
        # revert to defaults
        assert 6380 == config['lower_port']
        assert 6449 == config['upper_port']
        # should have printed an error
        out, err = capsys.readouterr()
        assert ("Invalid port_range '%s', should be like '6380-6449'\n" % (port_range)) == err

    with_directory_contents(
        {
            DEFAULT_LOCAL_STATE_FILENAME: """
service_options:
  REDIS_URL:
    port_range: %s
         """ % port_range
        }, read_config)


def test_garbage_port_range(capsys):
    _read_invalid_port_range(capsys, "abcdefg")


def test_backward_port_range(capsys):
    _read_invalid_port_range(capsys, "100-99")


def test_non_integer_port_range(capsys):
    _read_invalid_port_range(capsys, "A-Z")


def test_zero_lower_port(capsys):
    _read_invalid_port_range(capsys, "0-1")


def test_zero_upper_port(capsys):
    _read_invalid_port_range(capsys, "1-0")


def test_set_config_values_as_strings():
    def set_config(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = _redis_requirement()
        provider = RedisProvider()
        provider.set_config_values_as_strings(requirement,
                                              dict(),
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(lower_port="6001"))
        config = provider.read_config(requirement, dict(), local_state, 'default', UserConfigOverrides())
        assert config['lower_port'] == 6001
        assert config['upper_port'] == 6449

        provider.set_config_values_as_strings(requirement,
                                              dict(),
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(upper_port="6700"))
        config2 = provider.read_config(requirement, dict(), local_state, 'default', UserConfigOverrides())
        assert config2['lower_port'] == 6001
        assert config2['upper_port'] == 6700

        provider.set_config_values_as_strings(requirement,
                                              dict(),
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(lower_port="5500",
                                                   upper_port="6800"))
        config2 = provider.read_config(requirement, dict(), local_state, 'default', UserConfigOverrides())
        assert config2['lower_port'] == 5500
        assert config2['upper_port'] == 6800

    with_directory_contents(dict(), set_config)


def _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch):
    can_connect_args = dict()

    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        can_connect_args['host'] = host
        can_connect_args['port'] = port
        can_connect_args['timeout_seconds'] = timeout_seconds
        return True

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)

    return can_connect_args


def test_prepare_redis_url_with_dict_in_variables_section(monkeypatch):
    can_connect_args = _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch)

    def prepare_redis_url(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert result
        assert dict(REDIS_URL="redis://localhost:6379",
                    PROJECT_DIR=project.directory_path) == strip_environ(result.environ)
        assert dict(host='localhost', port=6379, timeout_seconds=0.5) == can_connect_args

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, prepare_redis_url)


def _monkeypatch_can_connect_to_socket_on_nonstandard_port_only(monkeypatch, real_can_connect_to_socket):
    can_connect_args_list = []

    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        can_connect_args = dict()
        can_connect_args['host'] = host
        can_connect_args['port'] = port
        can_connect_args['timeout_seconds'] = timeout_seconds
        can_connect_args_list.append(can_connect_args)
        if port == 6379:
            return False
        else:
            return real_can_connect_to_socket(host, port, timeout_seconds)

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)

    return can_connect_args_list


def test_prepare_and_unprepare_local_redis_server(monkeypatch):
    # this test will fail if you don't have Redis installed, since
    # it actually starts it.
    if platform.system() == 'Windows':
        print("Cannot start redis-server on Windows")
        return

    from anaconda_project.requirements_registry.network_util import can_connect_to_socket as real_can_connect_to_socket

    can_connect_args_list = _monkeypatch_can_connect_to_socket_on_nonstandard_port_only(monkeypatch,
                                                                                        real_can_connect_to_socket)

    def start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert result

        local_state_file = LocalStateFile.load_for_directory(dirname)
        state = local_state_file.get_service_run_state('REDIS_URL')
        assert 'port' in state
        port = state['port']

        assert dict(REDIS_URL=("redis://localhost:" + str(port)),
                    PROJECT_DIR=project.directory_path) == strip_environ(result.environ)
        assert len(can_connect_args_list) >= 2

        servicedir = os.path.join(dirname, "services")
        redisdir = os.path.join(servicedir, "REDIS_URL")

        pidfile = os.path.join(redisdir, "redis.pid")
        logfile = os.path.join(redisdir, "redis.log")
        assert os.path.exists(pidfile)
        assert os.path.exists(logfile)

        assert real_can_connect_to_socket(host='localhost', port=port)

        # now clean it up
        status = unprepare(project, result)
        assert status

        assert not os.path.exists(pidfile)
        assert not os.path.exists(logfile)
        assert not os.path.exists(redisdir)
        assert not os.path.exists(servicedir)
        assert not real_can_connect_to_socket(host='localhost', port=port)

        local_state_file.load()
        assert dict() == local_state_file.get_service_run_state("REDIS_URL")

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, start_local_redis)


def test_prepare_and_unprepare_local_redis_server_with_failed_unprovide(monkeypatch):
    # this test will fail if you don't have Redis installed, since
    # it actually starts it.
    if platform.system() == 'Windows':
        print("Cannot start redis-server on Windows")
        return

    from anaconda_project.requirements_registry.network_util import can_connect_to_socket as real_can_connect_to_socket

    _monkeypatch_can_connect_to_socket_on_nonstandard_port_only(monkeypatch, real_can_connect_to_socket)

    def start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert result

        # now clean it up, but arrange for that to fail
        local_state_file = LocalStateFile.load_for_directory(dirname)
        local_state_file.set_service_run_state('REDIS_URL', {'shutdown_commands': [['false']]})
        local_state_file.save()
        status = unprepare(project, result)
        assert not status
        assert status.status_description == 'Shutdown commands failed for REDIS_URL.'

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, start_local_redis)


def test_prepare_and_unprepare_two_local_redis_servers_with_failed_unprovide(monkeypatch):
    # this test will fail if you don't have Redis installed, since
    # it actually starts it.
    if platform.system() == 'Windows':
        print("Cannot start redis-server on Windows")
        return

    from anaconda_project.requirements_registry.network_util import can_connect_to_socket as real_can_connect_to_socket

    _monkeypatch_can_connect_to_socket_on_nonstandard_port_only(monkeypatch, real_can_connect_to_socket)

    def start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert result

        # now clean it up, but arrange for that to double-fail
        local_state_file = LocalStateFile.load_for_directory(dirname)
        local_state_file.set_service_run_state('REDIS_URL', {'shutdown_commands': [['false']]})
        local_state_file.set_service_run_state('REDIS_URL_2', {'shutdown_commands': [['false']]})
        local_state_file.save()
        status = unprepare(project, result)
        assert not status
        assert status.status_description == 'Failed to clean up REDIS_URL, REDIS_URL_2.'

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
  REDIS_URL_2: redis
"""}, start_local_redis)


def test_prepare_local_redis_server_twice_reuses(monkeypatch):
    # this test will fail if you don't have Redis installed, since
    # it actually starts it.
    if platform.system() == 'Windows':
        print("Cannot start redis-server on Windows")
        return

    from anaconda_project.requirements_registry.network_util import can_connect_to_socket as real_can_connect_to_socket

    can_connect_args_list = _monkeypatch_can_connect_to_socket_on_nonstandard_port_only(monkeypatch,
                                                                                        real_can_connect_to_socket)

    def start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert result
        assert 'REDIS_URL' in result.environ

        local_state_file = LocalStateFile.load_for_directory(dirname)
        state = local_state_file.get_service_run_state("REDIS_URL")
        assert 'port' in state
        port = state['port']

        assert dict(REDIS_URL=("redis://localhost:" + str(port)),
                    PROJECT_DIR=project.directory_path) == strip_environ(result.environ)
        assert len(can_connect_args_list) >= 2

        pidfile = os.path.join(dirname, "services/REDIS_URL/redis.pid")
        logfile = os.path.join(dirname, "services/REDIS_URL/redis.log")
        assert os.path.exists(pidfile)
        assert os.path.exists(logfile)

        assert real_can_connect_to_socket(host='localhost', port=port)

        # be sure we generate the config html that would use the old one
        requirement = _redis_requirement()
        status = requirement.check_status(result.environ, local_state_file, 'default', UserConfigOverrides())

        # now try again, and we should re-use the exact same server
        pidfile_mtime = os.path.getmtime(pidfile)
        with codecs.open(pidfile, 'r', 'utf-8') as file:
            pidfile_content = file.read()
        result2 = _prepare_printing_errors(project, environ=minimal_environ())
        assert result2

        # port should be the same, and set in the environment
        assert dict(REDIS_URL=("redis://localhost:" + str(port)),
                    PROJECT_DIR=project.directory_path) == strip_environ(result2.environ)

        # no new pid file
        assert pidfile_mtime == os.path.getmtime(pidfile)
        with codecs.open(pidfile, 'r', 'utf-8') as file:
            pidfile_content2 = file.read()
        assert pidfile_content == pidfile_content2

        # now clean it up
        status = unprepare(project, result2)
        assert status

        assert not os.path.exists(pidfile)
        assert not real_can_connect_to_socket(host='localhost', port=port)

        local_state_file.load()
        assert dict() == local_state_file.get_service_run_state("REDIS_URL")

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, start_local_redis)


def test_prepare_local_redis_server_times_out(monkeypatch, capsys):
    # this test will fail if you don't have Redis installed, since
    # it actually starts it.
    if platform.system() == 'Windows':
        print("Cannot start redis-server on Windows")
        return

    from anaconda_project.requirements_registry.network_util import can_connect_to_socket as real_can_connect_to_socket

    _monkeypatch_can_connect_to_socket_on_nonstandard_port_only(monkeypatch, real_can_connect_to_socket)

    def start_local_redis_and_time_out(dirname):
        project = project_no_dedicated_env(dirname)

        from time import sleep as real_sleep

        killed = {}

        def mock_sleep_kills_redis(seconds):
            # first time the Redis provider sleeps to wait for the
            # server to appear, we kill the server; after that
            # we make sleep into a no-op so we rapidly time out.
            if 'done' in killed:
                return

            pidfile = os.path.join(dirname, "services", "REDIS_URL", "redis.pid")
            count = 0
            while count < 15:
                if os.path.exists(pidfile):
                    break
                real_sleep(0.1)
                count = count + 1

            assert os.path.exists(pidfile)

            with codecs.open(pidfile, 'r', 'utf-8') as f:
                for line in f.readlines():
                    try:
                        import signal
                        os.kill(int(line.strip()), signal.SIGKILL)
                    except Exception:
                        pass

            # be sure it's gone
            real_sleep(0.1)
            killed['done'] = True

        monkeypatch.setattr('time.sleep', mock_sleep_kills_redis)

        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert not result

        out, err = capsys.readouterr()
        assert "redis-server started successfully, but we timed out trying to connect to it on port" in out
        assert "redis-server process failed or timed out, exited with code 0" in err

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, start_local_redis_and_time_out)


def _monkeypatch_can_connect_to_socket_always_succeeds_on_nonstandard(monkeypatch):
    can_connect_args_list = []

    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        can_connect_args = dict()
        can_connect_args['host'] = host
        can_connect_args['port'] = port
        can_connect_args['timeout_seconds'] = timeout_seconds
        can_connect_args_list.append(can_connect_args)
        return port != 6379

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)

    return can_connect_args_list


def test_fail_to_prepare_local_redis_server_no_port_available(monkeypatch, capsys):
    can_connect_args_list = _monkeypatch_can_connect_to_socket_always_succeeds_on_nonstandard(monkeypatch)

    def start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert not result
        assert 73 == len(can_connect_args_list)

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, start_local_redis)

    out, err = capsys.readouterr()
    assert "All ports from 6380 to 6449 were in use, could not start redis-server on one of them." in err
    assert "REDIS_URL" in err
    assert "missing requirement" in err
    assert "" == out


def test_do_not_start_local_redis_server_in_prod_mode(monkeypatch, capsys):
    can_connect_args_list = _monkeypatch_can_connect_to_socket_always_succeeds_on_nonstandard(monkeypatch)

    def no_start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ(), mode=provide.PROVIDE_MODE_PRODUCTION)
        assert not result
        assert 3 == len(can_connect_args_list)

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, no_start_local_redis)

    out, err = capsys.readouterr()
    assert "Could not connect to system default Redis." in err
    assert "REDIS_URL" in err
    assert "missing requirement" in err
    assert "" == out


def test_do_not_start_local_redis_server_in_check_mode(monkeypatch, capsys):
    can_connect_args_list = _monkeypatch_can_connect_to_socket_always_succeeds_on_nonstandard(monkeypatch)

    def no_start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ(), mode=provide.PROVIDE_MODE_CHECK)
        assert not result
        assert 3 == len(can_connect_args_list)

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, no_start_local_redis)

    out, err = capsys.readouterr()
    assert "Could not connect to system default Redis." in err
    assert "REDIS_URL" in err
    assert "missing requirement" in err
    assert "" == out


def _monkeypatch_can_connect_to_socket_always_fails(monkeypatch):
    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        return False

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)


def test_fail_to_prepare_local_redis_server_scope_system(monkeypatch, capsys):
    _monkeypatch_can_connect_to_socket_always_fails(monkeypatch)

    def check_no_autostart(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert not result

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
""",
         DEFAULT_LOCAL_STATE_FILENAME: """
service_options:
  REDIS_URL:
    scope: system
"""}, check_no_autostart)

    out, err = capsys.readouterr()
    assert out == ""
    assert err == (
        "Could not connect to system default Redis.\n" +
        "missing requirement to run this project: A running Redis server, located by a redis: URL set as REDIS_URL.\n" +
        "  Environment variable REDIS_URL is not set.\n")


def test_redis_server_configure_custom_port_range(monkeypatch, capsys):
    can_connect_args_list = _monkeypatch_can_connect_to_socket_always_succeeds_on_nonstandard(monkeypatch)

    def start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert not result
        assert 36 == len(can_connect_args_list)

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
    """,
         DEFAULT_LOCAL_STATE_FILENAME: """
service_options:
  REDIS_URL:
    port_range: 7389-7421
"""}, start_local_redis)

    out, err = capsys.readouterr()
    assert "All ports from 7389 to 7421 were in use, could not start redis-server on one of them." in err
    assert "REDIS_URL" in err
    assert "missing requirement" in err
    assert "" == out


def _fail_to_prepare_local_redis_server_exec_fails(monkeypatch, capsys, logfile_fail_mode):
    # this test will fail if you don't have Redis installed, since
    # it actually starts it.
    from anaconda_project.requirements_registry.network_util import can_connect_to_socket as real_can_connect_to_socket

    _monkeypatch_can_connect_to_socket_on_nonstandard_port_only(monkeypatch, real_can_connect_to_socket)

    def start_local_redis(dirname):
        logfile = os.path.join(dirname, "services/REDIS_URL/redis.log")

        from subprocess import Popen as real_Popen

        failscript = os.path.join(dirname, "fail.py")
        with codecs.open(failscript, 'w', 'utf-8') as file:
            file.write("""
from __future__ import print_function
import codecs
import sys
import os
print('It did not work stdout')
print('It did not work stderr', file=sys.stderr)
logfile = sys.argv[1]
fail_mode = sys.argv[2]
if fail_mode == 'no_logfile':
    pass
elif fail_mode == 'is_dir':
    os.makedirs(logfile)
else:
    with codecs.open(logfile, 'w', 'utf-8') as f:
        f.write('This is in the logfile')
sys.exit(1)
""")

        def mock_Popen(*args, **kwargs):
            if 'args' not in kwargs:
                # `pip list` goes through this codepath while redis launch
                # happens to specify args= as a kwarg
                assert 'pip' in args[0][0]
                return real_Popen(*args, **kwargs)
            kwargs['args'] = ['python', failscript, logfile, logfile_fail_mode]
            return real_Popen(*args, **kwargs)

        monkeypatch.setattr("subprocess.Popen", mock_Popen)

        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert not result

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, start_local_redis)

    # this doesn't capture "It did not work stdout" because
    # of some pytest detail I don't understand.
    out, err = capsys.readouterr()

    assert "REDIS_URL" in err
    assert "missing requirement" in err
    # the failed process writes this to stderr, but prepare() moves it to stdout
    assert "Starting " in out
    assert "It did not work stderr" in out
    if logfile_fail_mode == 'logfile_ok':
        assert "This is in the logfile" in out
    else:
        assert "This is in the logfile" not in out
    if logfile_fail_mode == 'is_dir':
        assert "Failed to read" in out
    else:
        assert "Failed to read" not in out


def test_fail_to_prepare_local_redis_server_exec_fails(monkeypatch, capsys):
    _fail_to_prepare_local_redis_server_exec_fails(monkeypatch, capsys, logfile_fail_mode='logfile_ok')


def test_fail_to_prepare_local_redis_server_exec_fails_no_logfile(monkeypatch, capsys):
    _fail_to_prepare_local_redis_server_exec_fails(monkeypatch, capsys, logfile_fail_mode='no_logfile')


def test_fail_to_prepare_local_redis_server_exec_fails_logfile_is_dir(monkeypatch, capsys):
    _fail_to_prepare_local_redis_server_exec_fails(monkeypatch, capsys, logfile_fail_mode='is_dir')


def test_fail_to_prepare_local_redis_server_not_on_path(monkeypatch, capsys):
    from anaconda_project.requirements_registry.network_util import can_connect_to_socket as real_can_connect_to_socket

    _monkeypatch_can_connect_to_socket_on_nonstandard_port_only(monkeypatch, real_can_connect_to_socket)

    def start_local_redis(dirname):
        from subprocess import Popen as real_Popen

        def mock_Popen(*args, **kwargs):
            if 'args' not in kwargs:
                # `pip list` goes through this codepath while redis launch
                # happens to specify args= as a kwarg
                assert 'pip' in args[0][0]
                return real_Popen(*args, **kwargs)
            kwargs['args'] = ['this-is-not-on-the-path']
            return real_Popen(*args, **kwargs)

        monkeypatch.setattr("subprocess.Popen", mock_Popen)

        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert not result

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, start_local_redis)

    # this doesn't capture "It did not work stdout" because
    # of some pytest detail I don't understand.
    out, err = capsys.readouterr()

    assert "REDIS_URL" in err
    assert "missing requirement" in err
    assert "Error executing redis-server: " in err


def test_set_scope_in_local_state(monkeypatch):
    can_connect_args = _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch)

    def prepare_after_setting_scope(dirname):
        local_state = LocalStateFile.load_for_directory(dirname)
        requirement = _redis_requirement()
        provider = RedisProvider()
        environ = minimal_environ()
        config = provider.read_config(requirement, environ, local_state, 'default', UserConfigOverrides())
        assert config['source'] == 'find_all'
        provider.set_config_values_as_strings(requirement,
                                              environ,
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(source='find_project'))
        config = provider.read_config(requirement, environ, local_state, 'default', UserConfigOverrides())
        assert config['source'] == 'find_project'
        provider.set_config_values_as_strings(requirement,
                                              environ,
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(source='find_all'))
        config = provider.read_config(requirement, environ, local_state, 'default', UserConfigOverrides())
        assert config['source'] == 'find_all'
        provider.set_config_values_as_strings(requirement,
                                              environ,
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(source='environ'))
        config = provider.read_config(requirement, environ, local_state, 'default', UserConfigOverrides())
        assert config['source'] == 'find_all'  # default if no env var set
        provider.set_config_values_as_strings(requirement,
                                              environ,
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(source='environ'))
        environ_with_redis_url = environ.copy()
        environ_with_redis_url['REDIS_URL'] = 'blah'
        config = provider.read_config(requirement, environ_with_redis_url, local_state, 'default',
                                      UserConfigOverrides())
        assert config['source'] == 'environ'  # default when the env var IS set

        # use local variable when env var not set
        provider.set_config_values_as_strings(requirement,
                                              environ,
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(source='variables',
                                                   value='foo'))
        config = provider.read_config(requirement, environ, local_state, 'default', UserConfigOverrides())
        assert config['source'] == 'variables'
        assert config['value'] == 'foo'

        # use local variable when env var _is_ set
        provider.set_config_values_as_strings(requirement,
                                              environ_with_redis_url,
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(source='variables',
                                                   value='foo'))
        config = provider.read_config(requirement, environ, local_state, 'default', UserConfigOverrides())
        assert config['source'] == 'variables'
        assert config['value'] == 'foo'

        # set to use system, which should override using the local state
        provider.set_config_values_as_strings(requirement,
                                              environ,
                                              local_state,
                                              'default',
                                              UserConfigOverrides(),
                                              dict(source='find_system'))
        config = provider.read_config(requirement, environ, local_state, 'default', UserConfigOverrides())
        assert config['source'] == 'find_system'

        project = project_no_dedicated_env(dirname)
        result = _prepare_printing_errors(project, environ=minimal_environ())
        assert result
        assert dict(REDIS_URL="redis://localhost:6379",
                    PROJECT_DIR=project.directory_path) == strip_environ(result.environ)
        assert dict(host='localhost', port=6379, timeout_seconds=0.5) == can_connect_args

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""}, prepare_after_setting_scope)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os
import sys

from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.project_lock_file import DEFAULT_PROJECT_LOCK_FILENAME
from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents


def _monkeypatch_pwd(monkeypatch, dirname):
    from os.path import abspath as real_abspath

    def mock_abspath(path):
        if path == ".":
            return dirname
        else:
            return real_abspath(path)

    monkeypatch.setattr('os.path.abspath', mock_abspath)


def _monkeypatch_isatty(monkeypatch, value):
    def mock_isatty():
        return value

    # python 2 can throw a "readonly" error if you try to patch sys.stdin.isatty itself
    monkeypatch.setattr('anaconda_project.internal.cli.console_utils.stdin_is_interactive', mock_isatty)


def test_init_in_pwd(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'init'])
        assert code == 0

        assert os.path.isfile(os.path.join(dirname, DEFAULT_PROJECT_FILENAME))
        assert not os.path.isfile(os.path.join(dirname, DEFAULT_PROJECT_LOCK_FILENAME))

        out, err = capsys.readouterr()
        assert ("Project configuration is in %s\n" % (os.path.join(dirname, DEFAULT_PROJECT_FILENAME))) == out
        assert '' == err

    with_directory_contents(dict(), check)


def test_init_create_directory(capsys, monkeypatch):
    _monkeypatch_isatty(monkeypatch, True)

    def mock_input_yes(prompt):
        sys.stdout.write(prompt)
        return "yes"

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils._input', mock_input_yes)

    def check(dirname):
        subdir = os.path.join(dirname, "foo")

        code = _parse_args_and_run_subcommand(['anaconda-project', 'init', '--directory', subdir])
        assert code == 0

        assert os.path.isfile(os.path.join(subdir, DEFAULT_PROJECT_FILENAME))
        assert not os.path.isfile(os.path.join(dirname, DEFAULT_PROJECT_LOCK_FILENAME))

        out, err = capsys.readouterr()

        assert ("Create directory '%s'? Project configuration is in %s\n" %
                (subdir, os.path.join(subdir, DEFAULT_PROJECT_FILENAME))) == out
        assert '' == err

    with_directory_contents(dict(), check)


def test_init_do_not_create_directory(capsys, monkeypatch):
    _monkeypatch_isatty(monkeypatch, True)

    def mock_input_no(prompt):
        sys.stdout.write(prompt)
        return "no"

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils._input', mock_input_no)

    def check(dirname):
        subdir = os.path.join(dirname, "foo")

        code = _parse_args_and_run_subcommand(['anaconda-project', 'init', '--directory', subdir])
        assert code == 1

        assert not os.path.isfile(os.path.join(subdir, DEFAULT_PROJECT_FILENAME))
        assert not os.path.isdir(subdir)

        out, err = capsys.readouterr()
        assert ("Create directory '%s'? " % subdir) == out
        assert ("Project directory '%s' does not exist.\nUnable to load the project.\n" % subdir) == err

    with_directory_contents(dict(), check)


def test_init_do_not_create_directory_not_interactive(capsys, monkeypatch):
    _monkeypatch_isatty(monkeypatch, False)

    def mock_input(prompt):
        raise RuntimeError("This should not have been called")

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils._input', mock_input)

    def check(dirname):
        subdir = os.path.join(dirname, "foo")

        code = _parse_args_and_run_subcommand(['anaconda-project', 'init', '--directory', subdir])
        assert code == 1

        assert not os.path.isfile(os.path.join(subdir, DEFAULT_PROJECT_FILENAME))
        assert not os.path.isdir(subdir)

        out, err = capsys.readouterr()
        assert '' == out
        assert ("Project directory '%s' does not exist.\nUnable to load the project.\n" % subdir) == err

    with_directory_contents(dict(), check)


def test_init_create_directory_not_interactive_with_yes(capsys, monkeypatch):
    _monkeypatch_isatty(monkeypatch, False)

    def mock_input(prompt):
        raise RuntimeError("This should not have been called")

    monkeypatch.setattr('anaconda_project.internal.cli.console_utils._input', mock_input)

    def check(dirname):
        subdir = os.path.join(dirname, "foo")

        code = _parse_args_and_run_subcommand(['anaconda-project', 'init', '--yes', '--directory', subdir])
        assert code == 0

        assert os.path.isfile(os.path.join(subdir, DEFAULT_PROJECT_FILENAME))
        assert not os.path.isfile(os.path.join(dirname, DEFAULT_PROJECT_LOCK_FILENAME))
        assert os.path.isdir(subdir)

        out, err = capsys.readouterr()
        assert ("Project configuration is in %s\n" % (os.path.join(subdir, DEFAULT_PROJECT_FILENAME))) == out
        assert '' == err

    with_directory_contents(dict(), check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from tornado import httpclient
from tornado import gen

import anaconda_project.internal.makedirs as makedirs
import anaconda_project.internal.rename as rename

import os
import hashlib


class FileDownloader(object):
    def __init__(self, url, filename, hash_algorithm=None):
        """Downloader for the given url to the given filename, computing the given hash.

        hash_algorithm is the name of a hash function in hashlib
        """
        self._url = url
        self._filename = filename
        self._hash_algorithm = hash_algorithm
        self._hash = None
        self._client = None
        self._errors = []

    @gen.coroutine
    def run(self, io_loop):
        """Run the download on the given io_loop."""
        assert self._client is None

        dirname = os.path.dirname(self._filename)
        try:
            makedirs.makedirs_ok_if_exists(dirname)
        except Exception as e:
            self._errors.append("Could not create directory '%s': %s" % (dirname, e))
            raise gen.Return(None)

        if self._hash_algorithm is not None:
            hasher = getattr(hashlib, self._hash_algorithm)()
        self._client = httpclient.AsyncHTTPClient(
            io_loop=io_loop,
            max_clients=1,
            # without this we buffer a huge amount
            # of stuff and then call the streaming_callback
            # once.
            max_buffer_size=1024 * 1024,
            # without this we 599 on large downloads
            max_body_size=100 * 1024 * 1024 * 1024,
            force_instance=True)

        tmp_filename = self._filename + ".part"
        try:
            _file = open(tmp_filename, 'wb')
        except EnvironmentError as e:
            self._errors.append("Failed to open %s: %s" % (tmp_filename, e))
            raise gen.Return(None)

        def cleanup_tmp():
            try:
                _file.close()
                # future: we could save it in order to try
                # resuming a failed download midstream, but
                # pointless until the download code above
                # knows how to resume.
                os.remove(tmp_filename)
            except EnvironmentError:
                pass

        def writer(chunk):
            if len(self._errors) > 0:
                return

            if self._hash_algorithm is not None:
                hasher.update(chunk)

            try:
                _file.write(chunk)
            except EnvironmentError as e:
                # we can't actually throw this error or Tornado freaks out, so instead
                # we ignore all future chunks once we have an error, which does mean
                # we continue to download bytes that we don't use. yuck.
                self._errors.append("Failed to write to %s: %s" % (tmp_filename, e))

        try:
            timeout_in_seconds = 60 * 10  # pretty long because we could be dealing with huge files
            request = httpclient.HTTPRequest(url=self._url,
                                             streaming_callback=writer,
                                             request_timeout=timeout_in_seconds)
            try:
                response = yield self._client.fetch(request)
            except Exception as e:
                self._errors.append("Failed download to %s: %s" % (self._filename, str(e)))
                raise gen.Return(None)

            # assert fetch() was supposed to throw the error, not leave it here unthrown
            assert response.error is None

            if len(self._errors) == 0:
                try:
                    _file.close()  # be sure tmp_filename is flushed
                    rename.rename_over_existing(tmp_filename, self._filename)
                except EnvironmentError as e:
                    self._errors.append("Failed to rename %s to %s: %s" % (tmp_filename, self._filename, str(e)))

            if len(self._errors) == 0 and self._hash_algorithm is not None:
                self._hash = hasher.hexdigest()

            raise gen.Return(response)
        finally:
            cleanup_tmp()

    @property
    def hash(self):
        """Hash of the downloaded file if we succeeded in downloading it, None if we failed."""
        return self._hash

    @property
    def errors(self):
        """List of errors if we failed to download, empty list if we succeeded."""
        return self._errors



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Command-line-specific project prepare utilities."""
from __future__ import absolute_import, print_function

from anaconda_project import prepare
from anaconda_project import project_ops
from anaconda_project.requirements_registry.requirement import EnvVarRequirement
from anaconda_project.requirements_registry.requirements.conda_env import CondaEnvRequirement

from anaconda_project.provide import (PROVIDE_MODE_PRODUCTION, PROVIDE_MODE_DEVELOPMENT, PROVIDE_MODE_CHECK)

import anaconda_project.internal.cli.console_utils as console_utils

# these UI_MODE_ strings are used as values for command line options, so they are user-visible

# ASK_QUESTIONS mode is supposed to ask about default actions too,
# like whether to start servers.  It isn't implemented yet.
UI_MODE_TEXT_ASK_QUESTIONS = "ask"
UI_MODE_TEXT_DEVELOPMENT_DEFAULTS_OR_ASK = "development_defaults_or_ask"
UI_MODE_TEXT_ASSUME_YES_PRODUCTION = "production_defaults"
UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT = "development_defaults"
UI_MODE_TEXT_ASSUME_NO = "check"

_all_ui_modes = (UI_MODE_TEXT_ASK_QUESTIONS, UI_MODE_TEXT_DEVELOPMENT_DEFAULTS_OR_ASK,
                 UI_MODE_TEXT_ASSUME_YES_PRODUCTION, UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT, UI_MODE_TEXT_ASSUME_NO)


def _interactively_fix_missing_variables(project, result):
    """Return True if we need to re-prepare."""
    if project.problems:
        return False

    if not console_utils.stdin_is_interactive():
        return False

    # We don't ask the user to manually enter CONDA_PREFIX
    # (CondaEnvRequirement) because it's a bizarre/confusing
    # thing to ask.
    can_ask_about = [status
                     for status in result.statuses
                     if (not status and isinstance(status.requirement, EnvVarRequirement) and not isinstance(
                         status.requirement, CondaEnvRequirement))]

    if can_ask_about:
        print("(Use Ctrl+C to quit.)")

    start_over = False
    values = dict()
    for status in can_ask_about:
        reply = console_utils.console_input("Value for " + status.requirement.env_var + ": ",
                                            encrypted=status.requirement.encrypted)
        if reply is None:
            return False  # EOF
        reply = reply.strip()
        if reply == '':
            start_over = True
            break
        values[status.requirement.env_var] = reply

    if len(values) > 0:
        status = project_ops.set_variables(project, result.env_spec_name, values.items(), result)
        if status:
            return True
        else:
            console_utils.print_status_errors(status)
            return False
    else:
        return start_over


def prepare_with_ui_mode_printing_errors(project,
                                         environ=None,
                                         ui_mode=UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT,
                                         env_spec_name=None,
                                         command_name=None,
                                         command=None,
                                         extra_command_args=None):
    """Perform all steps needed to get a project ready to execute.

    This may need to ask the user questions, may start services,
    run scripts, load configuration, install packages... it can do
    anything. Expect side effects.

    Args:
        project (Project): the project
        environ (dict): the environment to prepare (None to use os.environ)
        ui_mode (str): one of ``UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT``,
                       ``UI_MODE_TEXT_ASSUME_YES_PRODUCTION``, ``UI_MODE_TEXT_ASSUME_NO``
        env_spec_name (str): the environment spec name to require, or None for default
        command_name (str): command name to use or None for default
        command (ProjectCommand): a command object or None
        extra_command_args (list of str): extra args for the command we prepare

    Returns:
        a ``PrepareResult`` instance

    """
    assert ui_mode in _all_ui_modes  # the arg parser should have guaranteed this

    ask = False
    if ui_mode == UI_MODE_TEXT_ASSUME_YES_PRODUCTION:
        provide_mode = PROVIDE_MODE_PRODUCTION
    elif ui_mode == UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT:
        provide_mode = PROVIDE_MODE_DEVELOPMENT
    elif ui_mode == UI_MODE_TEXT_ASSUME_NO:
        provide_mode = PROVIDE_MODE_CHECK
    elif ui_mode == UI_MODE_TEXT_DEVELOPMENT_DEFAULTS_OR_ASK:
        provide_mode = PROVIDE_MODE_DEVELOPMENT
        ask = True

    # We might implement this by using
    # Provider.read_config/Provider.set_config_values_as_strings
    # or some new version of those; dig the old ui_server.py out
    # of git history to see how we used those methods to implement
    # an interactive HTML UI. read_config/set_config_values still
    # exist on Provider in case they are useful to implement this.
    assert ui_mode != UI_MODE_TEXT_ASK_QUESTIONS  # Not implemented yet

    # TODO: this could let you fix the suggestions if they are fixable.
    # (Note that we fix fatal problems in project_load.py, but we only
    #  display suggestions when we do a manual prepare, run, etc.)
    suggestions = project.suggestions
    if len(suggestions) > 0:
        print("Potential issues with this project:")
        for suggestion in project.suggestions:
            print("  * " + suggestion)
        print("")

    environ = None
    while True:
        result = prepare.prepare_without_interaction(project,
                                                     environ,
                                                     mode=provide_mode,
                                                     env_spec_name=env_spec_name,
                                                     command_name=command_name,
                                                     command=command,
                                                     extra_command_args=extra_command_args)

        if result.failed:
            if ask and _interactively_fix_missing_variables(project, result):
                environ = result.environ
                continue  # re-prepare, building on our previous environ

        # if we didn't continue, quit.
        break

    return result



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os

from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents_completing_project_file
from anaconda_project.project import Project
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME


class FakeRequirementStatus(object):
    def __init__(self, success, status_description):
        self.status_description = status_description
        self.success = success
        self.logs = ["This is a log message."]
        self.errors = []
        if not success:
            self.errors.append("This is an error message.")

    def __bool__(self):
        return self.success

    def __nonzero__(self):
        return self.success


def _monkeypatch_pwd(monkeypatch, dirname):
    from os.path import abspath as real_abspath

    def mock_abspath(path):
        if path == ".":
            return dirname
        else:
            return real_abspath(path)

    monkeypatch.setattr('os.path.abspath', mock_abspath)


def _monkeypatch_add_download(monkeypatch, result):
    params = {'args': (), 'kwargs': {}}

    def mock_add_download(*args, **kwargs):
        params['args'] = args
        params['kwargs'] = kwargs
        return result

    monkeypatch.setattr("anaconda_project.project_ops.add_download", mock_add_download)
    return params


def _test_add_download(capsys, monkeypatch, command):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        params = _monkeypatch_add_download(monkeypatch,
                                           FakeRequirementStatus(success=True,
                                                                 status_description='File downloaded.'))

        code = _parse_args_and_run_subcommand(command)
        assert code == 0

        out, err = capsys.readouterr()
        assert ('File downloaded.\n' + 'Added http://localhost:123456 to the project file.\n') == out
        assert '' == err
        return params

    return with_directory_contents_completing_project_file(dict(), check)


def test_add_download(capsys, monkeypatch):
    called_params = _test_add_download(capsys, monkeypatch,
                                       ['anaconda-project', 'add-download', 'MYDATA', 'http://localhost:123456'])
    assert len(called_params['args']) == 1
    expected_kwargs = {
        'env_spec_name': None,
        'env_var': 'MYDATA',
        'filename': None,
        'url': 'http://localhost:123456',
        'hash_algorithm': None,
        'hash_value': None
    }
    assert called_params['kwargs'] == expected_kwargs


def test_add_download_with_filename(capsys, monkeypatch):
    called_params = _test_add_download(
        capsys, monkeypatch,
        ['anaconda-project', 'add-download', 'MYDATA', 'http://localhost:123456', '--filename', 'foo'])
    assert len(called_params['args']) == 1
    expected_kwargs = {
        'env_spec_name': None,
        'env_var': 'MYDATA',
        'filename': 'foo',
        'url': 'http://localhost:123456',
        'hash_algorithm': None,
        'hash_value': None
    }
    assert called_params['kwargs'] == expected_kwargs


def test_add_download_with_checksum(capsys, monkeypatch):
    called_params = _test_add_download(capsys, monkeypatch, [
        'anaconda-project', 'add-download', 'MYDATA', 'http://localhost:123456', '--hash-algorithm', 'md5',
        '--hash-value', 'foo'
    ])
    assert len(called_params['args']) == 1
    expected_kwargs = {
        'env_spec_name': None,
        'env_var': 'MYDATA',
        'hash_algorithm': 'md5',
        'hash_value': 'foo',
        'url': 'http://localhost:123456',
        'filename': None
    }
    assert called_params['kwargs'] == expected_kwargs


def _test_add_download_with_only_one_hash_param(capsys, monkeypatch, command):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        params = _monkeypatch_add_download(monkeypatch,
                                           FakeRequirementStatus(success=True,
                                                                 status_description='File downloaded.'))

        code = _parse_args_and_run_subcommand(command)
        assert code == 1

        out, err = capsys.readouterr()
        assert "Error: mutually dependant parameters: --hash-algorithm and --hash-value.\n" == err
        assert '' == out
        return params

    return with_directory_contents_completing_project_file(dict(), check)


def test_add_download_with_only_hash_algorithm(capsys, monkeypatch):
    called_params = _test_add_download_with_only_one_hash_param(capsys, monkeypatch, [
        'anaconda-project', 'add-download', 'MYDATA', 'http://localhost:123456', '--hash-algorithm', 'md5'
    ])
    assert len(called_params['args']) == 0
    assert called_params['kwargs'] == {}


def test_add_download_with_only_hash_value(capsys, monkeypatch):
    called_params = _test_add_download_with_only_one_hash_param(capsys, monkeypatch, [
        'anaconda-project', 'add-download', 'MYDATA', 'http://localhost:123456', '--hash-value', 'foo'
    ])
    assert len(called_params['args']) == 0
    assert called_params['kwargs'] == {}


def _test_download_command_with_project_file_problems(capsys, monkeypatch, command):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        code = _parse_args_and_run_subcommand(command)
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert ('variables section contains wrong value type 42,' + ' should be dict or list of requirements\n' +
                'Unable to load the project.\n') in err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, check)


def test_add_download_command_with_project_file_problems(capsys, monkeypatch):
    _test_download_command_with_project_file_problems(
        capsys, monkeypatch, ['anaconda-project', 'add-download', 'MYDATA', 'http://localhost:123456'])


def test_remove_download_with_project_file_problems(capsys, monkeypatch):
    _test_download_command_with_project_file_problems(capsys, monkeypatch,
                                                      ['anaconda-project', 'remove-download', 'MYDATA'])


def test_remove_download(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-download', 'TEST_FILE'])
        project = Project(dirname)
        assert not project.downloads(project.default_env_spec_name)
        assert code == 0

        out, err = capsys.readouterr()
        filename = os.path.join(dirname, 'foo.tgz')
        assert ("Removed downloaded file %s.\nRemoved TEST_FILE from the project file.\n" % filename) == out
        assert '' == err

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: "downloads:\n  TEST_FILE: http://localhost/foo.tgz",
            'foo.tgz': 'data here'
        }, check)


def test_remove_download_dir(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-download', 'TEST_FILE'])
        project = Project(dirname)
        assert not project.downloads(project.default_env_spec_name)
        assert code == 0

        out, err = capsys.readouterr()
        filename = os.path.join(dirname, 'foo')
        assert ("Removed downloaded file %s.\nRemoved TEST_FILE from the project file.\n" % filename) == out
        assert '' == err

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: "downloads:\n  TEST_FILE: http://localhost/foo.zip",
            'foo/data.txt': 'data here'
        }, check)


def test_remove_download_file_error(capsys, monkeypatch):
    def check(dirname):
        from os import remove as real_remove
        _monkeypatch_pwd(monkeypatch, dirname)

        test_filename = os.path.join(dirname, 'foo.tgz')

        # only allow mock to have side effect once
        # later, when cleaning up TEST directory, allow removal
        mock_called = []

        def mock_remove(arg, *args, **kwargs):
            if arg == test_filename and not mock_called:
                mock_called.append(True)
                raise Exception('Error')
            return real_remove(arg, *args, **kwargs)

        monkeypatch.setattr('os.remove', mock_remove)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-download', 'TEST_FILE'])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert "Failed to remove {}: Error.\n".format(test_filename) == err

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: "downloads:\n  TEST_FILE: http://localhost/foo.tgz",
            'foo.tgz': 'data here'
        }, check)


def test_remove_download_directory_error(capsys, monkeypatch):
    def check(dirname):
        from shutil import rmtree as real_rmtree
        _monkeypatch_pwd(monkeypatch, dirname)

        test_filename = os.path.join(dirname, 'foo')

        # only allow mock to have side effect once
        # later, when cleaning up directory, allow removal
        mock_called = []

        def mock_remove(path, ignore_errors=False, onerror=None):
            if path == test_filename and not mock_called:
                mock_called.append(True)
                raise Exception('Error')
            return real_rmtree(path, ignore_errors, onerror)

        monkeypatch.setattr('shutil.rmtree', mock_remove)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-download', 'TEST_FILE'])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert "Failed to remove {}: Error.\n".format(test_filename) == err

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: "downloads:\n  TEST_FILE: http://localhost/foo.zip",
            'foo/data.txt': 'data here'
        }, check)


def test_remove_download_missing_var(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-download', 'TEST_FILE'])
        project = Project(dirname)
        assert not project.downloads(project.default_env_spec_name)
        assert code == 1

        out, err = capsys.readouterr()
        assert ("Download requirement: TEST_FILE not found.\n") == err
        assert '' == out

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: 'variables:\n  foo: {default: bar}'}, check)


def test_list_downloads_with_project_file_problems(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-downloads'])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert ('variables section contains wrong value type 42,' + ' should be dict or list of requirements\n' +
                'Unable to load the project.\n') in err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, check)


def test_list_empty_downloads(capsys, monkeypatch):
    def check_list_empty(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-downloads'])

        assert code == 0
        out, err = capsys.readouterr()
        expected_out = "No downloads found in project.\n"
        assert out == expected_out

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check_list_empty)


def test_list_downloads(capsys, monkeypatch):
    def check_list_not_empty(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-downloads'])
        assert code == 0
        out, err = capsys.readouterr()

        expected_out = """
Downloads for project: {dirname}

Name   Description
====   ===========
test   A downloaded file which is referenced by test.
train  A downloaded file which is referenced by train.
""".format(dirname=dirname).strip() + "\n"
        assert out == expected_out

    with_directory_contents_completing_project_file(
        {DEFAULT_PROJECT_FILENAME: ('downloads:\n'
                                    '  test: http://localhost:8000/test.tgz\n'
                                    '  train: http://localhost:8000/train.tgz\n')}, check_list_not_empty)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Utilities for implementing console interaction."""
from __future__ import absolute_import, print_function

import getpass
import sys

_PY2 = sys.version_info[0] == 2


def print_project_problems(project):
    """Print project problems to stderr, and return True if there were problems."""
    if project.problems:
        for problem in project.problems:
            print(problem, file=sys.stderr)
        print("Unable to load the project.", file=sys.stderr)
        return True
    else:
        return False


def print_status_errors(status):
    """Print out status description to stderr."""
    assert status is not None
    # don't print status.errors because we will have done that
    # already in streaming fashion from our Frontend.
    print(status.status_description, file=sys.stderr)


def format_names_and_descriptions(objects, name_attr='name', description_attr='description'):
    """Format a table with names on the left and descriptions on the right."""
    pairs = []
    for o in objects:
        name = getattr(o, name_attr)
        description = getattr(o, description_attr)
        pairs.append((name, description))

    # deterministic order
    pairs = sorted(pairs, key=lambda p: p[0])

    # add headers if there's anything in the list
    if len(pairs) > 0:
        pairs = [("Name", "Description"), ("====", "===========")] + pairs

    # format as table
    max_name_len = 0
    for pair in pairs:
        max_name_len = max(len(pair[0]), max_name_len)
    table = ""
    for pair in pairs:
        if pair[0] == pair[1]:
            # skip useless description
            line = pair[0]
        else:
            line = pair[0].ljust(max_name_len) + "  " + pair[1]
        table = table + line + "\n"
    if table == "":
        return "\n"
    else:
        return table


def print_names_and_descriptions(objects, name_attr='name', description_attr='description'):
    """Print a table with names on the left and descriptions on the right."""
    output = format_names_and_descriptions(objects, name_attr=name_attr, description_attr=description_attr)
    # we chop the newline off
    print(output[:-1])


def stdin_is_interactive():
    """True if stdin is a tty."""
    return sys.stdin.isatty()


# this "_input" wrapper exists to let us mock "input" because
# pytest makes it pesky to mock builtin functions that vary across
# python versions.  Python 2 has "input" and "raw_input" where
# "input" is eval(raw_input()).  Python 3 renames "raw_input" to
# "input".
def _input(prompt):
    if _PY2:  # pragma: no cover
        return raw_input(prompt)  # flake8: noqa # pragma: no cover (py2 only)
    else:
        return input(prompt)  # flake8: noqa # pragma: no cover


def console_input(prompt, encrypted=False):
    """Wrapper for raw_input (py2) and input (py3).

    Returns None on EOF.
    """
    try:
        if encrypted:
            return getpass.getpass(prompt)
        else:
            return _input(prompt)
    except EOFError:
        return None
    except KeyboardInterrupt:
        print("\nCanceling\n", file=sys.stderr)
        sys.exit(1)


def console_ask_yes_or_no(prompt, default):
    """Ask a yes or no question, returning a bool.

    Returns default if not on a console or EOF.
    """
    if not stdin_is_interactive():
        return default

    # show the "(enter y or n)" clutter only the second time around
    extra = ""
    while True:
        reply = console_input(prompt + extra + " ")
        if reply is None:
            return default

        if len(reply) > 0:
            if reply[0] in ('y', 'Y'):
                return True
            elif reply[0] in ('n', 'N'):
                return False

        extra = " (enter y or n):"



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Project lock file loading and manipulation."""
from __future__ import absolute_import

import os
from collections import OrderedDict

try:
    # this is the conda-packaged version of ruamel.yaml which has the
    # module renamed
    import ruamel_yaml as ryaml
except ImportError:  # pragma: no cover
    # this is the upstream version
    import ruamel.yaml as ryaml  # pragma: no cover

from anaconda_project.yaml_file import YamlFile, _CommentedMap, _block_style_all_nodes

# these are in the order we'll use them if multiple are present
possible_project_lock_file_names = ("anaconda-project-lock.yml", "anaconda-project-lock.yaml")

DEFAULT_PROJECT_LOCK_FILENAME = possible_project_lock_file_names[0]


class ProjectLockFile(YamlFile):
    """Represents the ``anaconda-project-lock.yml`` file which describes locked package versions."""

    @classmethod
    def load_for_directory(cls, directory):
        """Load the project lock file from the given directory, even if it doesn't exist.

        If the directory has no project file, the loaded
        ``ProjectLockFile`` will be empty. It won't actually be
        created on disk unless you call ``save()``.

        If the file has syntax problems, this sets the
        ``corrupted`` and ``corrupted_error_message`` properties,
        and attempts to modify the file will raise an
        exception. If the project file has semantic problems, they
        are not detected by this class but are reported by the
        ``Project`` class.

        Args:
            directory (str): path to the project directory

        Returns:
            a new ``ProjectLockFile``

        """
        for name in possible_project_lock_file_names:
            path = os.path.join(directory, name)
            if os.path.isfile(path):
                return ProjectLockFile(path)
        return ProjectLockFile(os.path.join(directory, DEFAULT_PROJECT_LOCK_FILENAME))

    def __init__(self, filename):
        """Construct a ``ProjectLockFile`` with the given filename.

        It's easier to use ``ProjectLockFile.load_for_directory()`` in most cases.

        If the file has syntax problems, this sets the
        ``corrupted`` and ``corrupted_error_message`` properties,
        and attempts to modify the file will raise an
        exception. If the project file has semantic problems, they
        are not detected by this class but are reported by the
        ``Project`` class.

        Args:
            filename (str): path to the project file
        """
        super(ProjectLockFile, self).__init__(filename)

    def _default_content(self):
        header = (
            "This is an Anaconda project lock file.\n" +
            "The lock file locks down exact versions of all your dependencies.\n" + "\n" +
            "In most cases, this file is automatically maintained by the `anaconda-project` command or GUI tools.\n" +
            "It's best to keep this file in revision control (such as git or svn).\n" +
            "The file is in YAML format, please see http://www.yaml.org/start.html for more.\n")
        sections = OrderedDict()

        sections['locking_enabled'] = ("Set to false to ignore locked versions.")
        sections['env_specs'] = ("A key goes in here for each env spec.\n")

        # we make a big string and then parse it because I can't figure out the
        # ruamel.yaml API to insert comments in front of map keys.
        def comment_out(comment):
            return ("# " + "\n# ".join(comment.split("\n")) + "\n").replace("# \n", "#\n")

        to_parse = comment_out(header)
        for section_name, comment in sections.items():
            # future: this is if/else is silly, we should be
            # assigning these bodies up above when we assign the
            # comments.
            if section_name in ('env_specs', ):
                section_body = "  {}"
            elif section_name in ('locking_enabled', ):
                section_body = " false"
            to_parse = to_parse + "\n#\n" + comment_out(comment) + section_name + ":\n" + section_body + "\n\n\n"

        as_json = ryaml.load(to_parse, Loader=ryaml.RoundTripLoader)

        return as_json

    def _save_default_content(self):
        # We don't want to save empty lock files.
        return False

    def _set_lock_set_hash(self, env_spec_name, env_spec_hash):
        """Library-internal method."""
        self.set_value(['env_specs', env_spec_name, 'env_spec_hash'], env_spec_hash)

    def _set_lock_set(self, env_spec_name, lock_set, all_names):
        """Library-internal method."""
        assert env_spec_name is not None
        assert lock_set is not None
        assert all_names is not None
        assert env_spec_name in all_names

        # if all locking is disabled, switch to individually
        # disabling each env spec so we can enable the one
        # we care about here.
        for_all = self.get_value(['locking_enabled'], True)
        if not for_all:
            self.set_value(['locking_enabled'], True)

            new_env_specs = self.get_value(['env_specs'], None)
            if new_env_specs is None:
                new_env_specs = _CommentedMap()
                _block_style_all_nodes(new_env_specs)
                self.set_value(['env_specs'], new_env_specs)

            for name in all_names:
                if name == env_spec_name:
                    continue
                single_env = new_env_specs.get(name, None)
                if single_env is None:
                    single_env = _CommentedMap()
                    _block_style_all_nodes(single_env)
                    new_env_specs[name] = single_env

                single_env['locked'] = False

        # now set up the one env
        as_json = lock_set.to_json()
        self.set_value(['env_specs', env_spec_name], as_json)

    def _disable_locking(self, env_spec_name):
        """Library-internal method."""
        if env_spec_name is None:
            self.set_value(['locking_enabled'], False)
            self.unset_value(['env_specs'])
        else:
            self.set_value(['env_specs', env_spec_name, 'locked'], False)
            self.unset_value(['env_specs', env_spec_name, 'packages'])
            self.unset_value(['env_specs', env_spec_name, 'platforms'])
